#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# PySNMP MIB module PYSNMP-MIB (https://www.pysnmp.com/pysnmp)
# ASN.1 source http://mibs.pysnmp.com:80/asn1/PYSNMP-MIB
# Produced by pysmi-0.1.3 at Mon Apr 17 11:46:02 2017
# On host grommit.local platform Darwin version 16.4.0 by user ilya
# Using Python version 3.4.2 (v3.4.2:ab2c023a9432, Oct  5 2014, 20:42:22)
#
Integer, OctetString, ObjectIdentifier = mibBuilder.import_symbols(
    "ASN1", "Integer", "OctetString", "ObjectIdentifier"
)
(NamedValues,) = mibBuilder.import_symbols("ASN1-ENUMERATION", "NamedValues")
(
    SingleValueConstraint,
    ValueRangeConstraint,
    ConstraintsIntersection,
    ValueSizeConstraint,
    ConstraintsUnion,
) = mibBuilder.import_symbols(
    "ASN1-REFINEMENT",
    "SingleValueConstraint",
    "ValueRangeConstraint",
    "ConstraintsIntersection",
    "ValueSizeConstraint",
    "ConstraintsUnion",
)
NotificationGroup, ModuleCompliance = mibBuilder.import_symbols(
    "SNMPv2-CONF", "NotificationGroup", "ModuleCompliance"
)
(
    ModuleIdentity,
    iso,
    MibScalar,
    MibTable,
    MibTableRow,
    MibTableColumn,
    Gauge32,
    NotificationType,
    IpAddress,
    MibIdentifier,
    Unsigned32,
    Counter32,
    ObjectIdentity,
    Counter64,
    Bits,
    Integer32,
    enterprises,
    TimeTicks,
) = mibBuilder.import_symbols(
    "SNMPv2-SMI",
    "ModuleIdentity",
    "iso",
    "MibScalar",
    "MibTable",
    "MibTableRow",
    "MibTableColumn",
    "Gauge32",
    "NotificationType",
    "IpAddress",
    "MibIdentifier",
    "Unsigned32",
    "Counter32",
    "ObjectIdentity",
    "Counter64",
    "Bits",
    "Integer32",
    "enterprises",
    "TimeTicks",
)
TextualConvention, DisplayString = mibBuilder.import_symbols(
    "SNMPv2-TC", "TextualConvention", "DisplayString"
)
pysnmp = ModuleIdentity((1, 3, 6, 1, 4, 1, 20408))
if mibBuilder.loadTexts:
    pysnmp.setRevisions(
        (
            "2017-04-14 00:00",
            "2005-05-14 00:00",
        )
    )
if mibBuilder.loadTexts:
    pysnmp.setLastUpdated("201704140000Z")
if mibBuilder.loadTexts:
    pysnmp.setOrganization("The PySNMP Project")
if mibBuilder.loadTexts:
    pysnmp.setContactInfo(
        "E-mail: LeXtudio Inc. <support@lextudio.com> GitHub: https://github.com/lextudio/pysnmp"
    )
if mibBuilder.loadTexts:
    pysnmp.setDescription("PySNMP top-level MIB tree infrastructure")
pysnmpObjects = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 1))
pysnmpExamples = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 2))
pysnmpEnumerations = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3))
pysnmpModuleIDs = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3, 1))
pysnmpAgentOIDs = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3, 2))
pysnmpDomains = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3, 3))
pysnmpExperimental = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 9999))
pysnmpNotificationPrefix = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 4))
pysnmpNotifications = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 4, 0))
pysnmpNotificationObjects = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 4, 1))
pysnmpConformance = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 5))
pysnmpCompliances = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 5, 1))
pysnmpGroups = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 5, 2))
mibBuilder.export_symbols(
    "PYSNMP-MIB",
    pysnmpCompliances=pysnmpCompliances,
    pysnmpObjects=pysnmpObjects,
    pysnmpNotificationPrefix=pysnmpNotificationPrefix,
    pysnmpModuleIDs=pysnmpModuleIDs,
    pysnmpGroups=pysnmpGroups,
    pysnmpNotificationObjects=pysnmpNotificationObjects,
    pysnmp=pysnmp,
    pysnmpExperimental=pysnmpExperimental,
    pysnmpNotifications=pysnmpNotifications,
    PYSNMP_MODULE_ID=pysnmp,
    pysnmpEnumerations=pysnmpEnumerations,
    pysnmpDomains=pysnmpDomains,
    pysnmpAgentOIDs=pysnmpAgentOIDs,
    pysnmpConformance=pysnmpConformance,
    pysnmpExamples=pysnmpExamples,
)



"""
This module is part of the pysnmp library and is necessary to make the 'rfc3826/priv' directory a package.

Filepath: /Users/lextm/pysnmp.com/pysnmp/pysnmp/proto/secmod/rfc3826/priv/__init__.py
"""



"""
Multiple SNMP Engines
+++++++++++++++++++++

Run multiple SNMP Engines each with a complete Command Responder.
Bind each SNMP Engine to a dedicated network transport endpoint:

* IPv4/UDP, listening at 127.0.0.1:161
* IPv4/UDP, listening at 127.0.0.2:161

Each Command Responder will respond to SNMP GET/SET/GETNEXT/GETBULK
queries with the following options:

* SNMPv3
* with USM user 'usr-md5-des', auth: MD5, priv DES
* allow read access to SNMPv2-MIB objects (1.3.6)
* allow write access to SNMPv2-MIB objects (1.3.6.1.2.1)

The following Net-SNMP commands will walk the first and the second
Agent respectively:

| $ snmpwalk -Ob -v3 -u usr-md5-des -l authPriv -A authkey1 -X privkey1 127.0.0.1 usmUserEntry
| $ snmpwalk -Ob -v3 -u usr-md5-des -l authPriv -A authkey1 -X privkey1 127.0.0.2 usmUserEntry

Notice differently configured snmpEngineId's in usmUserEntry columns.

"""  #
from pysnmp.entity import engine, config
from pysnmp.entity.rfc3413 import cmdrsp, context
from pysnmp.proto import rfc1902
from pysnmp.carrier.asyncio.dispatch import AsyncioDispatcher
from pysnmp.carrier.asyncio.dgram import udp

# Configuration parameters for each of SNMP Engines
snmpEngineInfo = (
    ("0102030405060708", udp.DOMAIN_NAME + (0,), ("127.0.0.1", 161)),
    ("0807060504030201", udp.DOMAIN_NAME + (1,), ("127.0.0.2", 161)),
)

# Instantiate the single transport dispatcher object
transportDispatcher = AsyncioDispatcher()

# Setup a custom data routing function to select snmpEngine by transportDomain
transportDispatcher.register_routing_callback(lambda td, t, d: td)

# Instantiate and configure SNMP Engines
for snmpEngineId, transportDomain, transportAddress in snmpEngineInfo:
    # Create SNMP engine with specific engineID
    snmpEngine = engine.SnmpEngine(rfc1902.OctetString(hexValue=snmpEngineId))

    # Register SNMP Engine object with transport dispatcher. Request incoming
    # data from specific transport endpoint to be funneled to this SNMP Engine.
    snmpEngine.register_transport_dispatcher(transportDispatcher, transportDomain)

    # Transport setup

    # UDP over IPv4
    config.add_transport(
        snmpEngine,
        transportDomain,
        udp.UdpTransport().open_server_mode(transportAddress),
    )

    # SNMPv3/USM setup

    # user: usr-md5-des, auth: MD5, priv DES
    config.add_v3_user(
        snmpEngine,
        "usr-md5-des",
        config.USM_AUTH_HMAC96_MD5,
        "authkey1",
        config.USM_PRIV_CBC56_DES,
        "privkey1",
    )

    # Allow full MIB access for this user / securityModels at VACM
    config.add_vacm_user(
        snmpEngine, 3, "usr-md5-des", "authPriv", (1, 3, 6), (1, 3, 6, 1, 2, 1)
    )

    # Get default SNMP context this SNMP engine serves
    snmpContext = context.SnmpContext(snmpEngine)

    # Register SNMP Applications at the SNMP engine for particular SNMP context
    cmdrsp.GetCommandResponder(snmpEngine, snmpContext)
    cmdrsp.SetCommandResponder(snmpEngine, snmpContext)
    cmdrsp.NextCommandResponder(snmpEngine, snmpContext)
    cmdrsp.BulkCommandResponder(snmpEngine, snmpContext)

# Register an imaginary never-ending job to keep I/O dispatcher running forever
transportDispatcher.job_started(1)

# Run I/O dispatcher which would receive queries and send responses
try:
    print("This program needs to run as root/administrator to monitor port 161.")
    print("Started. Press Ctrl-C to stop")
    transportDispatcher.run_dispatcher()

except KeyboardInterrupt:
    print("Shutting down...")

finally:
    transportDispatcher.close_dispatcher()



"""
PDU var-binds to MIB objects
++++++++++++++++++++++++++++

This script explains how Python application could turn SNMP PDU
variable-bindings into MIB objects or the other way around.

The code that configures MIB compiler is similar to what
happens inside the pysnmp.hlapi API.
"""  #
from pysnmp.smi import builder, view, compiler, rfc1902

# Assemble MIB browser
mibBuilder = builder.MibBuilder()
mibViewController = view.MibViewController(mibBuilder)
compiler.add_mib_compiler(
    mibBuilder,
    sources=["file:///usr/share/snmp/mibs", "https://mibs.pysnmp.com/asn1/@mib@"],
)

# Pre-load MIB modules we expect to work with
mibBuilder.load_modules("SNMPv2-MIB", "SNMP-COMMUNITY-MIB")

# This is what we can get in TRAP PDU
varBinds = [
    ("1.3.6.1.2.1.1.3.0", 12345),
    ("1.3.6.1.6.3.1.1.4.1.0", "1.3.6.1.6.3.1.1.5.2"),
    ("1.3.6.1.6.3.18.1.3.0", "0.0.0.0"),
    ("1.3.6.1.6.3.18.1.4.0", ""),
    ("1.3.6.1.6.3.1.1.4.3.0", "1.3.6.1.4.1.20408.4.1.1.2"),
    ("1.3.6.1.2.1.1.1.0", "my system"),
]

# Run var-binds through MIB resolver
# You may want to catch and ignore resolution errors here
varBinds = [
    rfc1902.ObjectType(rfc1902.ObjectIdentity(x[0]), x[1]).resolve_with_mib(
        mibViewController
    )
    for x in varBinds
]

for varBind in varBinds:
    print(varBind.prettyPrint())



from datetime import datetime
import os
import subprocess
from pathlib import Path
import sys
from inspect import isawaitable

from sphinx_polyversion.api import apply_overrides
from sphinx_polyversion.driver import DefaultDriver
from sphinx_polyversion.git import Git, GitRef, GitRefType, file_predicate, refs_by_type
from sphinx_polyversion.pyvenv import VirtualPythonEnvironment, Poetry
from sphinx_polyversion.sphinx import SphinxBuilder

#: Regex matching the branches to build docs for
BRANCH_REGEX = r"rel-.*"  # intentionally broken to skip branches

#: Regex matching the tags to build docs for
TAG_REGEX = r"v7.1|v6.2"

#: Output dir relative to project root
OUTPUT_DIR = "docs/build"

#: Source directory
SOURCE_DIR = "docs/"

#: Arguments to pass to `uv pip install`
UV_ARGS = "-e .[dev]"

#: Arguments to pass to Poetry install (for older versions)
POETRY_ARGS = "--sync"

#: Arguments to pass to `sphinx-build`
SPHINX_ARGS = "-a -v"

#: Mock data used for building local version
MOCK_DATA = {
    "revisions": [
        GitRef("v1.8.0", "", "", GitRefType.TAG, datetime.fromtimestamp(0)),
        GitRef("v1.9.3", "", "", GitRefType.TAG, datetime.fromtimestamp(1)),
        GitRef("v1.10.5", "", "", GitRefType.TAG, datetime.fromtimestamp(2)),
        GitRef("master", "", "", GitRefType.BRANCH, datetime.fromtimestamp(3)),
        GitRef("dev", "", "", GitRefType.BRANCH, datetime.fromtimestamp(4)),
        GitRef("some-feature", "", "", GitRefType.BRANCH, datetime.fromtimestamp(5)),
    ],
    "current": GitRef("local", "", "", GitRefType.TAG, datetime.fromtimestamp(6)),
}
MOCK = False


# Custom UV environment provider
class Uv(VirtualPythonEnvironment):
    """Python environment using uv."""

    def __init__(
        self,
        path: Path,
        name: str,
        venv: "str | Path",
        *,
        creator=None,
        env=None,
        args=None,
    ):
        """Initialize UV environment provider.

        Args:
            path: Path to the project
            name: Name of the environment
            venv: Path to the virtual environment
            creator: Function to create the virtual environment
            env: Environment variables to set
            args: Arguments to pass to `uv pip install`
        """
        super().__init__(
            path,
            name,
            venv,
            creator=creator,
            env=env,
        )
        self.args = args if args is not None else []

    @classmethod
    def factory(cls, args=None):
        """Create a factory that produces UV environments.

        Args:
            args: Arguments to pass to `uv pip install`

        Returns:
            A factory function that creates UV environments
        """
        args_list = args if isinstance(args, list) else (args.split() if args else [])

        def create_env(path, name):
            venv = path / ".venv"
            return cls(path, name, venv, args=args_list)

        return create_env

    async def create_venv(self) -> None:
        """Create the virtual python environment.

        Override of the VirtualPythonEnvironment.create_venv method to use uv.
        """
        if self._creator:
            result = self._creator(self.venv)
            if isawaitable(result):
                await result
        elif not self.venv.exists():
            # Create virtual environment
            path = self.venv
            subprocess.check_call([sys.executable, "-m", "venv", str(path)])

            # Get the pip path inside the venv
            if os.name == "nt":  # Windows
                pip_path = path / "Scripts" / "pip"
                python_path = path / "Scripts" / "python"
            else:  # Unix/macOS
                pip_path = path / "bin" / "pip"
                python_path = path / "bin" / "python"

            # Install uv in the virtual environment
            subprocess.check_call([str(pip_path), "install", "uv"])

            # Get the uv path inside the venv
            if os.name == "nt":  # Windows
                uv_path = path / "Scripts" / "uv"
            else:  # Unix/macOS
                uv_path = path / "bin" / "uv"

            # Install the project and dependencies using uv
            cmd = [str(uv_path), "pip", "install"] + self.args
            subprocess.check_call(cmd)

            # Save the python path for the sphinx-builder to use
            self.python_path = python_path

    async def __aenter__(self):
        """Set up the environment."""
        await super().__aenter__()
        return self

    def install_package(self, path):
        """Install a package in the environment.

        Args:
            path: Path to the package to install
        """
        # This is required by VirtualPythonEnvironment but we've already
        # installed everything in create_venv(), so this is a no-op for uv
        pass


# Function to choose the right environment provider based on version
def get_env_provider(version):
    """Get the appropriate environment provider based on the version.

    Args:
        version: Version string

    Returns:
        An environment provider factory function
    """
    if version.startswith("v7.1"):
        return Uv.factory(args=UV_ARGS)
    else:
        return Poetry.factory(args=POETRY_ARGS.split())


#: Data passed to templates
def data(driver, rev, env):
    revisions = driver.targets
    branches, tags = refs_by_type(revisions)
    latest = max(tags or branches)
    return {
        "current": rev,
        "tags": tags,
        "branches": branches,
        "revisions": revisions,
        "latest": latest,
    }


def root_data(driver):
    revisions = driver.builds
    branches, tags = refs_by_type(revisions)
    latest = max(tags or branches)
    return {"revisions": revisions, "latest": latest}


# Custom handler for different versions
class VersionBasedDriver(DefaultDriver):
    async def init_environment(self, path, rev):
        """Initialize the environment for a specific revision.

        Args:
            path: Path to create the environment at
            rev: Git reference

        Returns:
            The prepared environment
        """
        # Create the environment provider for the specific version
        env_provider = get_env_provider(rev.name)
        env = env_provider(path=path, name=rev.name)

        # Create the environment
        await env.__aenter__()

        return env


# Load overrides read from commandline to global scope
apply_overrides(globals())
# Determine repository root directory
root = Git.root(Path(__file__).parent)

# Setup driver and run it
src = Path(SOURCE_DIR)
VersionBasedDriver(
    root,
    OUTPUT_DIR,
    vcs=Git(
        branch_regex=BRANCH_REGEX,
        tag_regex=TAG_REGEX,
        buffer_size=1 * 10**9,  # 1 GB
        predicate=file_predicate([src]),  # exclude refs without source dir
    ),
    builder=SphinxBuilder(src / "source", args=SPHINX_ARGS.split()),
    env=None,  # We'll select the environment provider based on version
    template_dir=root / src / "templates",
    static_dir=root / src / "static",
    data_factory=data,
    root_data_factory=root_data,
    mock=MOCK_DATA,
).run(MOCK)



"""
Send crafted TRAP PDU
+++++++++++++++++++++

Initialize TRAP PDU and pass it over to unified SNMPv3 message processing
framework for further treatment.

* SNMPv2c
* with community name 'public'
* over IPv4/UDP
* send TRAP notification
* initialize TRAP PDU with the following var-binds:
  1.3.6.1.2.1.1.3.0 = 123
  1.3.6.1.6.3.1.1.4.1.0 = 1.3.6.1.6.3.1.1.5.1

Functionally similar to:

| $ snmptrap -v2c -c public demo.pysnmp.com 0 1.3.6.1.6.3.1.1.5.1

"""  #

from pysnmp.entity import engine, config
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.entity.rfc3413 import ntforg
from pysnmp.proto.api import v2c

# Create SNMP engine instance
snmpEngine = engine.SnmpEngine()

# SecurityName <-> CommunityName mapping
config.add_v1_system(snmpEngine, "my-area", "public")

# Specify security settings per SecurityName (SNMPv2c -> 1)
config.add_target_parameters(snmpEngine, "my-creds", "my-area", "noAuthNoPriv", 1)

# Setup transport endpoint and bind it with security settings yielding
# a target name
config.add_transport(
    snmpEngine, udp.DOMAIN_NAME, udp.UdpAsyncioTransport().open_client_mode()
)

# Create named target
config.add_target_address(
    snmpEngine, "my-nms", udp.DOMAIN_NAME, ("127.0.0.1", 162), "my-creds"
)

# *** SNMP engine configuration is complete by this line ***

# Create SNMP v2c TRAP PDU with defaults
trapPDU = v2c.TrapPDU()
v2c.apiTrapPDU.set_defaults(trapPDU)

# Set custom var-binds to TRAP PDU
v2c.apiTrapPDU.set_varbinds(
    trapPDU,
    [
        # sysUpTime
        (v2c.ObjectIdentifier("1.3.6.1.2.1.1.3.0"), v2c.TimeTicks(123)),
        # snmpTrapPDU
        (
            (1, 3, 6, 1, 6, 3, 1, 1, 4, 1, 0),
            v2c.ObjectIdentifier((1, 3, 6, 1, 6, 3, 1, 1, 5, 1)),
        ),
    ],
)

# Create Notification Originator App instance.
ntfOrg = ntforg.NotificationOriginator()


# Error/confirmation receiver
# noinspection PyUnusedLocal,PyUnusedLocal,PyUnusedLocal,PyUnusedLocal,PyUnusedLocal
def cbFun(
    snmpEngine,
    sendRequestHandle,
    errorIndication,
    errorStatus,
    errorIndex,
    varBinds,
    cbCtx,
):
    print(
        "Notification {}, status - {}".format(
            sendRequestHandle, errorIndication and errorIndication or "delivered"
        )
    )


# Build and submit notification message to dispatcher
ntfOrg.send_pdu(
    snmpEngine,
    # Notification targets
    "my-nms",  # target address
    None,
    "",  # contextEngineId, contextName
    trapPDU,
    cbFun,
)

print("Notification is scheduled to be sent")

# Run I/O dispatcher which would send pending message and process response
snmpEngine.open_dispatcher()



"""
SNMPv1
++++++

* with SNMPv1, community 'public'
* over IPv4/UDP
* to an Agent at 127.0.0.1:161
* for an OID in tuple form

This script performs similar to the following Net-SNMP command:

| $ snmpget -v1 -c public -ObentU 127.0.0.1 1.3.6.1.2.1.1.1.0

"""  #
from pysnmp.entity import engine, config
from pysnmp.carrier.asyncio.dgram import udp
from pysnmp.entity.rfc3413 import cmdgen

# Create SNMP engine instance
snmpEngine = engine.SnmpEngine()

#
# SNMPv1 setup
#

# SecurityName <-> CommunityName mapping
config.add_v1_system(snmpEngine, "my-area", "public")

# Specify security settings per SecurityName (SNMPv1 - 0, SNMPv2c - 1)
config.add_target_parameters(snmpEngine, "my-creds", "my-area", "noAuthNoPriv", 0)

#
# Setup transport endpoint and bind it with security settings yielding
# a target name
#

# UDP/IPv4
config.add_transport(
    snmpEngine, udp.DOMAIN_NAME, udp.UdpAsyncioTransport().open_client_mode()
)
config.add_target_address(
    snmpEngine, "my-router", udp.DOMAIN_NAME, ("127.0.0.1", 161), "my-creds"
)


# Error/response receiver
# noinspection PyUnusedLocal,PyUnusedLocal,PyUnusedLocal
def cbFun(
    snmpEngine,
    sendRequestHandle,
    errorIndication,
    errorStatus,
    errorIndex,
    varBinds,
    cbCtx,
):
    if errorIndication:
        print(errorIndication)
    # SNMPv1 response may contain noSuchName error *and* SNMPv2c exception,
    # so we ignore noSuchName error here
    elif errorStatus and errorStatus != 2:
        print(
            f"{errorStatus.prettyPrint()} at {varBinds[int(errorIndex) - 1][0] if errorIndex else '?'}"
        )
    else:
        for oid, val in varBinds:
            print(f"{oid.prettyPrint()} = {val.prettyPrint()}")


# Prepare and send a request message
cmdgen.GetCommandGenerator().send_varbinds(
    snmpEngine,
    "my-router",
    None,
    "",  # contextEngineId, contextName
    [((1, 3, 6, 1, 2, 1, 1, 1, 0), None)],
    cbFun,
)

# Run I/O dispatcher which would send pending queries and process responses
snmpEngine.open_dispatcher(3)

snmpEngine.close_dispatcher()



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
#
# Copyright (C) 2014, Zebra Technologies
# Authors: Matt Hooks <me@matthooks.com>
#          Zachary Lorusso <zlorusso@gmail.com>
#
# Copyright (C) 2024, LeXtudio Inc. <support@lextudio.com>
#
# License: https://www.pysnmp.com/pysnmp/license.html
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
#
"""
DgramAsyncioProtocol is a base class for asyncio datagram transport, designed to be used with AsyncioDispatcher.

Attributes:
    SOCK_FAMILY (int): Socket family type.
    ADDRESS_TYPE (type): Type of the transport address.
    transport (asyncio.DatagramTransport | None): The asyncio transport instance.
    loop (asyncio.AbstractEventLoop): The event loop instance.

Methods:
    __init__(sock=None, sockMap=None, loop=None):
        Initializes the datagram protocol object for asyncio.

    datagram_received(datagram, transportAddress):
        Processes incoming datagram.

    connection_made(transport):
        Prepares to send datagrams.

    connection_lost(exc):
        Cleans up after connection is lost.

    open_client_mode(iface=None, allow_broadcast=False):
        Opens client mode.

    open_server_mode(iface=None, sock=None):
        Opens server mode.

    close_transport():
        Closes the transport.

    send_message(outgoingMessage, transportAddress):
        Sends a message to the transport.

    normalize_address(transportAddress):
        Returns a transport address object.
"""
import asyncio
import sys
import traceback
import warnings
from socket import socket

from pysnmp import debug
from pysnmp.carrier import error
from pysnmp.carrier.asyncio.base import AbstractAsyncioTransport
from pysnmp.carrier.base import AbstractTransportAddress


class DgramAsyncioProtocol(asyncio.DatagramProtocol, AbstractAsyncioTransport):
    """Base Asyncio datagram Transport, to be used with AsyncioDispatcher."""

    SOCK_FAMILY: int = 0
    ADDRESS_TYPE: type
    transport: "asyncio.DatagramTransport | None" = None
    loop: asyncio.AbstractEventLoop

    def __init__(
        self, sock=None, sockMap=None, loop: "asyncio.AbstractEventLoop | None" = None
    ):
        """Create a datagram protocol object for asyncio."""
        self._writeQ = []
        self._lport = None
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop

    def datagram_received(self, datagram, transportAddress: AbstractTransportAddress):
        """Process incoming datagram."""
        if self._callback_function is None:
            raise error.CarrierError("Unable to call cbFun")
        else:
            self.loop.call_soon(
                self._callback_function, self, transportAddress, datagram
            )

    def connection_made(self, transport: asyncio.DatagramTransport):
        """Prepare to send datagrams."""
        self.transport = transport
        debug.logger & debug.FLAG_IO and debug.logger("connection_made: invoked")
        while self._writeQ:
            outgoingMessage, transportAddress = self._writeQ.pop(0)
            debug.logger & debug.FLAG_IO and debug.logger(
                "connection_made: transportAddress %r outgoingMessage %s"
                % (transportAddress, debug.hexdump(outgoingMessage))
            )
            try:
                self.transport.sendto(
                    outgoingMessage, self.normalize_address(transportAddress)  # type: ignore
                )
            except Exception:
                raise error.CarrierError(
                    ";".join(traceback.format_exception(*sys.exc_info()))
                )

    def connection_lost(self, exc):
        """Clean up after connection is lost."""
        debug.logger & debug.FLAG_IO and debug.logger("connection_lost: invoked")

    # AbstractAsyncioTransport API

    def open_client_mode(
        self, iface: "tuple[str, int] | None" = None, allow_broadcast: bool = False
    ):
        """Open client mode."""
        if self.loop.is_closed():
            raise error.CarrierError("Event loop is closed")

        try:
            c = self.loop.create_datagram_endpoint(
                lambda: self,
                local_addr=iface,
                family=self.SOCK_FAMILY,
                allow_broadcast=allow_broadcast,
            )
            # Avoid deprecation warning for asyncio.async()
            self._lport = asyncio.ensure_future(c)

        except Exception:
            raise error.CarrierError(
                ";".join(traceback.format_exception(*sys.exc_info()))
            )
        return self

    def open_server_mode(
        self, iface: "tuple[str, int] | None" = None, sock: "socket | None" = None
    ):
        """Open server mode."""
        if iface is None and sock is None:
            raise error.CarrierError("either iface or sock is required")

        if self.loop.is_closed():
            raise error.CarrierError("Event loop is closed")

        try:
            if sock:
                c = self.loop.create_datagram_endpoint(lambda: self, sock=sock)
            else:
                c = self.loop.create_datagram_endpoint(
                    lambda: self, local_addr=iface, family=self.SOCK_FAMILY
                )
            # Avoid deprecation warning for asyncio.async()
            self._lport = asyncio.ensure_future(c)
        except Exception:
            raise error.CarrierError(
                ";".join(traceback.format_exception(*sys.exc_info()))
            )
        return self

    def close_transport(self):
        """Close the transport."""
        if self._lport is not None:
            self._lport.cancel()
        if self.transport is not None:
            self.transport.close()
        AbstractAsyncioTransport.close_transport(self)

    def send_message(
        self,
        outgoingMessage,
        transportAddress: "AbstractTransportAddress | tuple[str, int]",
    ):
        """Send a message to the transport."""
        debug.logger & debug.FLAG_IO and debug.logger(
            "sendMessage: {} transportAddress {!r} outgoingMessage {}".format(
                (self.transport is None and "queuing" or "sending"),
                transportAddress,
                debug.hexdump(outgoingMessage),
            )
        )
        if self.transport is None:
            self._writeQ.append((outgoingMessage, transportAddress))
        else:
            try:
                self.transport.sendto(
                    outgoingMessage, self.normalize_address(transportAddress)  # type: ignore
                )
            except Exception:
                raise error.CarrierError(
                    ";".join(traceback.format_exception(*sys.exc_info()))
                )

    def normalize_address(
        self, transportAddress: "AbstractTransportAddress | tuple[str, int]"
    ):
        """Return a transport address object."""
        if not isinstance(transportAddress, self.ADDRESS_TYPE):
            transportAddress = self.ADDRESS_TYPE(transportAddress)
        return transportAddress

    # Old to new attribute mapping
    deprecated_attributes = {
        "openClientMode": "open_client_mode",
        "openServerMode": "open_server_mode",
        "closeTransport": "close_transport",
        "sendMessage": "send_message",
        "normalizeAddr": "normalize_address",
    }

    def __getattr__(self, attr: str):
        """Handle deprecated attributes."""
        if new_attr := self.deprecated_attributes.get(attr):
            warnings.warn(
                f"{attr} is deprecated. Please use {new_attr} instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            return getattr(self, new_attr)

        raise AttributeError(
            f"class '{self.__class__.__name__}' has no attribute '{attr}'"
        )



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from pyasn1.type import constraint, namedtype, namedval, tag, univ
from pysnmp.proto import rfc1902

__all__ = [
    "unSpecified",
    "EndOfMibView",
    "ReportPDU",
    "UnSpecified",
    "BulkPDU",
    "SNMPv2TrapPDU",
    "GetRequestPDU",
    "NoSuchObject",
    "GetNextRequestPDU",
    "GetBulkRequestPDU",
    "NoSuchInstance",
    "ResponsePDU",
    "noSuchObject",
    "InformRequestPDU",
    "endOfMibView",
    "SetRequestPDU",
    "noSuchInstance",
]

# Value reference -- max bindings in VarBindList
max_bindings = rfc1902.Integer(2147483647)

# Take SNMP exception values out of BindValue structure for convenience

UnSpecified = univ.Null
unSpecified = UnSpecified("")  # noqa: N816


class NoSuchObject(univ.Null):
    """No Such Object exception."""

    tagSet = univ.Null.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0x00)
    )

    def prettyPrint(self, scope=0):  # noqa: N802
        """Return a human-readable string representation of the exception."""
        return "No Such Object currently exists at this OID"


noSuchObject = NoSuchObject("")  # noqa: N816


class NoSuchInstance(univ.Null):
    """No Such Instance exception."""

    tagSet = univ.Null.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0x01)
    )

    def prettyPrint(self, scope=0):  # noqa: N802
        """Return a human-readable string representation of the exception."""
        return "No Such Instance currently exists at this OID"


noSuchInstance = NoSuchInstance("")  # noqa: N816


class EndOfMibView(univ.Null):
    """End of MIB View exception."""

    tagSet = univ.Null.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0x02)
    )

    def prettyPrint(self, scope=0):  # noqa: N802
        """Return a human-readable string representation of the exception."""
        return "No more variables left in this MIB View"


endOfMibView = EndOfMibView("")  # noqa: N816


# Made a separate class for better readability
class _BindValue(univ.Choice):
    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("value", rfc1902.ObjectSyntax()),
        namedtype.NamedType("unSpecified", unSpecified),
        namedtype.NamedType("noSuchObject", noSuchObject),
        namedtype.NamedType("noSuchInstance", noSuchInstance),
        namedtype.NamedType("endOfMibView", endOfMibView),
    )


class VarBind(univ.Sequence):
    """VarBind class."""

    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("name", rfc1902.ObjectName()),
        namedtype.NamedType("", _BindValue()),
    )


class VarBindList(univ.SequenceOf):
    """VarBindList class."""

    componentType = VarBind()  # noqa: N815
    subtypeSpec = univ.SequenceOf.subtypeSpec + constraint.ValueSizeConstraint(
        0, max_bindings
    )  # noqa: N815


errorStatus = univ.Integer(  # noqa: N816
    namedValues=namedval.NamedValues(
        ("noError", 0),
        ("tooBig", 1),
        ("noSuchName", 2),
        ("badValue", 3),
        ("readOnly", 4),
        ("genErr", 5),
        ("noAccess", 6),
        ("wrongType", 7),
        ("wrongLength", 8),
        ("wrongEncoding", 9),
        ("wrongValue", 10),
        ("noCreation", 11),
        ("inconsistentValue", 12),
        ("resourceUnavailable", 13),
        ("commitFailed", 14),
        ("undoFailed", 15),
        ("authorizationError", 16),
        ("notWritable", 17),
        ("inconsistentName", 18),
    )
)


# Base class for a non-bulk PDU
class PDU(univ.Sequence):
    """PDU class."""

    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("request-id", rfc1902.Integer32()),
        namedtype.NamedType("error-status", errorStatus),
        namedtype.NamedType(
            "error-index",
            univ.Integer().subtype(
                subtypeSpec=constraint.ValueRangeConstraint(0, max_bindings)
            ),
        ),
        namedtype.NamedType("variable-bindings", VarBindList()),
    )


nonRepeaters = univ.Integer().subtype(  # noqa: N816
    subtypeSpec=constraint.ValueRangeConstraint(0, max_bindings)
)
maxRepetitions = univ.Integer().subtype(  # noqa: N816
    subtypeSpec=constraint.ValueRangeConstraint(0, max_bindings)
)


# Base class for bulk PDU
class BulkPDU(univ.Sequence):
    """BulkPDU class."""

    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("request-id", rfc1902.Integer32()),
        namedtype.NamedType("non-repeaters", nonRepeaters),
        namedtype.NamedType("max-repetitions", maxRepetitions),
        namedtype.NamedType("variable-bindings", VarBindList()),
    )


class GetRequestPDU(PDU):
    """GetRequest PDU."""

    tagSet = PDU.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 0)
    )


class GetNextRequestPDU(PDU):
    """GetNextRequest PDU."""

    tagSet = PDU.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 1)
    )


class ResponsePDU(PDU):
    """Response PDU."""

    tagSet = PDU.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 2)
    )


class SetRequestPDU(PDU):
    """SetRequest PDU."""

    tagSet = PDU.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 3)
    )


class GetBulkRequestPDU(BulkPDU):
    """GetBulkRequest PDU."""

    tagSet = PDU.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 5)
    )


class InformRequestPDU(PDU):
    """InformRequest PDU."""

    tagSet = PDU.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 6)
    )


class SNMPv2TrapPDU(PDU):
    """SNMPv2Trap PDU."""

    tagSet = PDU.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 7)
    )


class ReportPDU(PDU):
    """Report PDU."""

    tagSet = PDU.tagSet.tagImplicitly(  # noqa: N815
        tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 8)
    )


class PDUs(univ.Choice):
    """PDUs class."""

    componentType = namedtype.NamedTypes(  # noqa: N815
        namedtype.NamedType("get-request", GetRequestPDU()),
        namedtype.NamedType("get-next-request", GetNextRequestPDU()),
        namedtype.NamedType("get-bulk-request", GetBulkRequestPDU()),
        namedtype.NamedType("response", ResponsePDU()),
        namedtype.NamedType("set-request", SetRequestPDU()),
        namedtype.NamedType("inform-request", InformRequestPDU()),
        namedtype.NamedType("snmpV2-trap", SNMPv2TrapPDU()),
        namedtype.NamedType("report", ReportPDU()),
    )



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# Copyright (C) 2014, Zebra Technologies
# Authors: Matt Hooks <me@matthooks.com>
#          Zachary Lorusso <zlorusso@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
#

import socket
import warnings
from typing import Tuple

from pysnmp.carrier.asyncio.dgram.base import DgramAsyncioProtocol
from pysnmp.carrier.base import AbstractTransportAddress

DOMAIN_NAME: Tuple[int, ...]
SNMP_UDP6_DOMAIN: Tuple[int, ...]
DOMAIN_NAME = SNMP_UDP_DOMAIN = (1, 3, 6, 1, 6, 1, 1)


class UdpTransportAddress(tuple, AbstractTransportAddress):
    """UDP transport address."""

    pass


class UdpAsyncioTransport(DgramAsyncioProtocol):
    """UDP async transport."""

    SOCK_FAMILY = socket.AF_INET
    ADDRESS_TYPE = UdpTransportAddress


UdpTransport = UdpAsyncioTransport

# Old to new attribute mapping
deprecated_attributes = {
    "domainName": "DOMAIN_NAME",
    "snmpUDPDomain": "SNMP_UDP_DOMAIN",
}


def __getattr__(attr: str):
    if new_attr := deprecated_attributes.get(attr):
        warnings.warn(
            f"{attr} is deprecated. Please use {new_attr} instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return globals()[new_attr]
    raise AttributeError(f"module '{__name__}' has no attribute '{attr}'")



import pytest

from pysnmp.hlapi.v3arch.asyncio import *
from pysnmp.proto.errind import RequestTimedOut
from tests.agent_context import AGENT_PORT, AgentContextManager


@pytest.mark.asyncio
@pytest.mark.parametrize("num_bulk", [1, 2, 50])
async def test_v2c_bulk(num_bulk):
    async with AgentContextManager():
        with SnmpEngine() as snmpEngine:
            errorIndication, errorStatus, errorIndex, varBinds = await bulk_cmd(
                snmpEngine,
                CommunityData("public"),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ContextData(),
                0,
                num_bulk,
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
                retries=0,
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == num_bulk
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysObjectID.0"
            if num_bulk > 1:
                assert varBinds[1][0].prettyPrint() == "SNMPv2-MIB::sysUpTime.0"
            if num_bulk > 2:
                assert varBinds[2][0].prettyPrint() == "SNMPv2-MIB::sysContact.0"


@pytest.mark.asyncio
async def test_v2c_bulk_non_exist():
    with SnmpEngine() as snmpEngine:
        errorIndication, errorStatus, errorIndex, varBinds = await bulk_cmd(
            snmpEngine,
            CommunityData("public"),
            await UdpTransportTarget.create(
                ("localhost", AGENT_PORT), timeout=0.5, retries=0
            ),
            ContextData(),
            0,
            1,
            ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
        )

        assert isinstance(errorIndication, RequestTimedOut)


@pytest.mark.asyncio
async def test_v2c_bulk_multiple_input():
    mib_objects = [
        ObjectType(ObjectIdentity("SNMPv2-MIB", "sysContact")),
        ObjectType(ObjectIdentity("SNMPv2-MIB", "sysORIndex")),
        ObjectType(ObjectIdentity("SNMPv2-MIB", "sysORDescr")),
    ]
    async with AgentContextManager():
        with SnmpEngine() as snmpEngine:
            errorIndication, errorStatus, errorIndex, varBinds = await bulk_cmd(
                snmpEngine,
                CommunityData("public"),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ContextData(),
                1,
                2,
                *mib_objects,
                retries=0
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 5


# snmpbulkget -v2c -c public -C n1 -C r2 localhost 1.3.6.1.2.1.1.4 1.3.6.1.2.1.1.9.1.1 1.3.6.1.2.1.1.9.1.3



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# Copyright (C) 2014, Zebra Technologies
# Authors: Matt Hooks <me@matthooks.com>
#          Zachary Lorusso <zlorusso@gmail.com>
# Modified by Ilya Etingof <ilya@snmplabs.com>
#
# Copyright (C) 2024, LeXtudio Inc. <support@lextudio.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
#
import asyncio
import warnings
from typing import AsyncGenerator

from pysnmp.entity.engine import SnmpEngine
from pysnmp.entity.rfc3413 import cmdgen
from pysnmp.hlapi import varbinds
from pysnmp.hlapi.v3arch.asyncio.auth import CommunityData, UsmUserData
from pysnmp.hlapi.v3arch.asyncio.context import ContextData
from pysnmp.hlapi.v3arch.asyncio.lcd import CommandGeneratorLcdConfigurator
from pysnmp.hlapi.v3arch.asyncio.transport import AbstractTransportTarget
from pysnmp.proto import errind
from pysnmp.proto.rfc1902 import Integer32, Null
from pysnmp.proto.rfc1905 import EndOfMibView, endOfMibView
from pysnmp.smi.rfc1902 import ObjectType


__all__ = [
    "get_cmd",
    "next_cmd",
    "set_cmd",
    "bulk_cmd",
    "walk_cmd",
    "bulk_walk_cmd",
    "is_end_of_mib",
]

VB_PROCESSOR = varbinds.CommandGeneratorVarBinds()
LCD = CommandGeneratorLcdConfigurator()
is_end_of_mib = varbinds.is_end_of_mib


async def get_cmd(
    snmpEngine: SnmpEngine,
    authData: "CommunityData | UsmUserData",
    transportTarget: AbstractTransportTarget,
    contextData: ContextData,
    *varBinds: ObjectType,
    **options,
) -> "tuple[errind.ErrorIndication, Integer32 | int, Integer32 | int, tuple[ObjectType, ...]]":
    r"""Creates a generator to perform SNMP GET query.

    When iterator gets advanced by :py:mod:`asyncio` main loop,
    SNMP GET request is send (:RFC:`1905#section-4.2.1`).
    The iterator yields :py:class:`asyncio.get_running_loop().create_future()` which gets done whenever
    response arrives or error occurs.

    Parameters
    ----------
    snmpEngine : :py:class:`~pysnmp.hlapi.v3arch.asyncio.SnmpEngine`
        Class instance representing SNMP engine.

    authData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.CommunityData` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.UsmUserData`
        Class instance representing SNMP credentials.

    transportTarget : :py:class:`~pysnmp.hlapi.v3arch.asyncio.UdpTransportTarget` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.Udp6TransportTarget`
        Class instance representing transport type along with SNMP peer address.

    contextData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.ContextData`
        Class instance representing SNMP ContextEngineId and ContextName values.

    \*varBinds : :py:class:`~pysnmp.smi.rfc1902.ObjectType`
        One or more class instances representing MIB variables to place
        into SNMP request.

    Other Parameters
    ----------------
    \*\*options :
        Request options:

            * `lookupMib` - load MIB and resolve response MIB variables at
              the cost of slightly reduced performance. Default is `True`.

    Yields
    ------
    errorIndication : :py:class:`~pysnmp.proto.errind.ErrorIndication`
        True value indicates SNMP engine error.
    errorStatus : str
        True value indicates SNMP PDU error.
    errorIndex : int
        Non-zero value refers to `varBinds[errorIndex-1]`
    varBinds : tuple
        A sequence of :py:class:`~pysnmp.smi.rfc1902.ObjectType` class
        instances representing MIB variables returned in SNMP response.

    Raises
    ------
    PySnmpError
        Or its derivative indicating that an error occurred while
        performing SNMP operation.

    Examples
    --------
    >>> import asyncio
    >>> from pysnmp.hlapi.v3arch.asyncio import *
    >>>
    >>> async def run():
    ...     errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
    ...         SnmpEngine(),
    ...         CommunityData('public'),
    ...         await UdpTransportTarget.create(('demo.pysnmp.com', 161)),
    ...         ContextData(),
    ...         ObjectType(ObjectIdentity('SNMPv2-MIB', 'sysDescr', 0))
    ...     )
    ...     print(errorIndication, errorStatus, errorIndex, varBinds)
    >>>
    >>> asyncio.run(run())
    (None, 0, 0, [ObjectType(ObjectIdentity(ObjectName('1.3.6.1.2.1.1.1.0')), DisplayString('SunOS zeus.pysnmp.com 4.1.3_U1 1 sun4m'))])
    >>>

    """

    def __callback(
        snmpEngine: SnmpEngine,
        sendRequestHandle,
        errorIndication: errind.ErrorIndication,
        errorStatus: "Integer32 | int",
        errorIndex: "Integer32 | int",
        varBinds,
        cbCtx,
    ):
        lookupMib, future = cbCtx
        if future.cancelled():
            return
        try:
            varBindsUnmade = VB_PROCESSOR.unmake_varbinds(
                snmpEngine.cache, varBinds, lookupMib
            )
        except Exception as e:
            future.set_exception(e)
        else:
            future.set_result((
                errorIndication,
                errorStatus,
                errorIndex,
                varBindsUnmade,
            ))

    addrName, paramsName = LCD.configure(
        snmpEngine, authData, transportTarget, contextData.contextName
    )

    future = asyncio.get_running_loop().create_future()

    cmdgen.GetCommandGenerator().send_varbinds(
        snmpEngine,
        addrName,
        contextData.contextEngineId,
        contextData.contextName,
        VB_PROCESSOR.make_varbinds(snmpEngine.cache, varBinds),
        __callback,
        (options.get("lookupMib", True), future),
    )
    return await future


async def set_cmd(
    snmpEngine: SnmpEngine,
    authData: "CommunityData | UsmUserData",
    transportTarget: AbstractTransportTarget,
    contextData: ContextData,
    *varBinds: ObjectType,
    **options,
) -> "tuple[errind.ErrorIndication, Integer32 | int, Integer32 | int, tuple[ObjectType, ...]]":
    r"""Creates a generator to perform SNMP SET query.

    When iterator gets advanced by :py:mod:`asyncio` main loop,
    SNMP SET request is send (:RFC:`1905#section-4.2.5`).
    The iterator yields :py:class:`asyncio.get_running_loop().create_future()` which gets done whenever
    response arrives or error occurs.

    Parameters
    ----------
    snmpEngine : :py:class:`~pysnmp.hlapi.v3arch.asyncio.SnmpEngine`
        Class instance representing SNMP engine.

    authData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.CommunityData` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.UsmUserData`
        Class instance representing SNMP credentials.

    transportTarget : :py:class:`~pysnmp.hlapi.v3arch.asyncio.UdpTransportTarget` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.Udp6TransportTarget`
        Class instance representing transport type along with SNMP peer address.

    contextData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.ContextData`
        Class instance representing SNMP ContextEngineId and ContextName values.

    \*varBinds : :py:class:`~pysnmp.smi.rfc1902.ObjectType`
        One or more class instances representing MIB variables to place
        into SNMP request.

    Other Parameters
    ----------------
    \*\*options :
        Request options:

            * `lookupMib` - load MIB and resolve response MIB variables at
              the cost of slightly reduced performance. Default is `True`.

    Yields
    ------
    errorIndication : :py:class:`~pysnmp.proto.errind.ErrorIndication`
        True value indicates SNMP engine error.
    errorStatus : str
        True value indicates SNMP PDU error.
    errorIndex : int
        Non-zero value refers to `varBinds[errorIndex-1]`
    varBinds : tuple
        A sequence of :py:class:`~pysnmp.smi.rfc1902.ObjectType` class
        instances representing MIB variables returned in SNMP response.

    Raises
    ------
    PySnmpError
        Or its derivative indicating that an error occurred while
        performing SNMP operation.

    Examples
    --------
    >>> import asyncio
    >>> from pysnmp.hlapi.v3arch.asyncio import *
    >>>
    >>> async def run():
    ...     errorIndication, errorStatus, errorIndex, varBinds = await set_cmd(
    ...         SnmpEngine(),
    ...         CommunityData('public'),
    ...         await UdpTransportTarget.create(('demo.pysnmp.com', 161)),
    ...         ContextData(),
    ...         ObjectType(ObjectIdentity('SNMPv2-MIB', 'sysDescr', 0), 'Linux i386')
    ...     )
    ...     print(errorIndication, errorStatus, errorIndex, varBinds)
    >>>
    >>> asyncio.run(run())
    (None, 0, 0, [ObjectType(ObjectIdentity(ObjectName('1.3.6.1.2.1.1.1.0')), DisplayString('Linux i386'))])
    >>>

    """

    def __callback(
        snmpEngine: SnmpEngine,
        sendRequestHandle,
        errorIndication: errind.ErrorIndication,
        errorStatus: "Integer32 | int",
        errorIndex: "Integer32 | int",
        varBinds,
        cbCtx,
    ):
        lookupMib, future = cbCtx
        if future.cancelled():
            return
        try:
            varBindsUnmade = VB_PROCESSOR.unmake_varbinds(
                snmpEngine.cache, varBinds, lookupMib
            )
        except Exception as e:
            future.set_exception(e)
        else:
            future.set_result((
                errorIndication,
                errorStatus,
                errorIndex,
                varBindsUnmade,
            ))

    addrName, paramsName = LCD.configure(
        snmpEngine, authData, transportTarget, contextData.contextName
    )

    future = asyncio.get_running_loop().create_future()

    cmdgen.SetCommandGenerator().send_varbinds(
        snmpEngine,
        addrName,
        contextData.contextEngineId,
        contextData.contextName,
        VB_PROCESSOR.make_varbinds(snmpEngine.cache, varBinds),
        __callback,
        (options.get("lookupMib", True), future),
    )
    return await future


async def next_cmd(
    snmpEngine: SnmpEngine,
    authData: "CommunityData | UsmUserData",
    transportTarget: AbstractTransportTarget,
    contextData: ContextData,
    *varBinds: ObjectType,
    **options,
) -> "tuple[errind.ErrorIndication, Integer32 | str | int, Integer32 | int, tuple[ObjectType, ...]]":
    r"""Creates a generator to perform SNMP GETNEXT query.

    When iterator gets advanced by :py:mod:`asyncio` main loop,
    SNMP GETNEXT request is send (:RFC:`1905#section-4.2.2`).
    The iterator yields :py:class:`asyncio.get_running_loop().create_future()` which gets done whenever
    response arrives or error occurs.

    Parameters
    ----------
    snmpEngine : :py:class:`~pysnmp.hlapi.v3arch.asyncio.SnmpEngine`
        Class instance representing SNMP engine.

    authData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.CommunityData` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.UsmUserData`
        Class instance representing SNMP credentials.

    transportTarget : :py:class:`~pysnmp.hlapi.v3arch.asyncio.UdpTransportTarget` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.Udp6TransportTarget`
        Class instance representing transport type along with SNMP peer address.

    contextData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.ContextData`
        Class instance representing SNMP ContextEngineId and ContextName values.

    \*varBinds : :py:class:`~pysnmp.smi.rfc1902.ObjectType`
        One or more class instances representing MIB variables to place
        into SNMP request.

    Other Parameters
    ----------------
    \*\*options :
        Request options:

            * `lookupMib` - load MIB and resolve response MIB variables at
              the cost of slightly reduced performance. Default is `True`.
            * `ignoreNonIncreasingOid` - continue iteration even if response
              MIB variables (OIDs) are not greater then request MIB variables.
              Be aware that setting it to `True` may cause infinite loop between
              SNMP management and agent applications. Default is `False`.

    Yields
    ------
    errorIndication : :py:class:`~pysnmp.proto.errind.ErrorIndication`
        True value indicates SNMP engine error.
    errorStatus : str
        True value indicates SNMP PDU error.
    errorIndex : int
        Non-zero value refers to `varBinds[errorIndex-1]`
    varBinds : tuple
        A sequence of sequences (e.g. 2-D array) of
        :py:class:`~pysnmp.smi.rfc1902.ObjectType` class instances
        representing a table of MIB variables returned in SNMP response.
        Inner sequences represent table rows and ordered exactly the same
        as `varBinds` in request. Response to GETNEXT always contain
        a single row.

    Raises
    ------
    PySnmpError
        Or its derivative indicating that an error occurred while
        performing SNMP operation.

    Examples
    --------
    >>> import asyncio
    >>> from pysnmp.hlapi.v3arch.asyncio import *
    >>>
    >>> async def run():
    ...     errorIndication, errorStatus, errorIndex, varBinds = await next_cmd(
    ...         SnmpEngine(),
    ...         CommunityData('public'),
    ...         await UdpTransportTarget.create(('demo.pysnmp.com', 161)),
    ...         ContextData(),
    ...         ObjectType(ObjectIdentity('SNMPv2-MIB', 'system'))
    ...     )
    ...     print(errorIndication, errorStatus, errorIndex, varBinds)
    >>>
    >>> asyncio.run(run())
    (None, 0, 0, [[ObjectType(ObjectIdentity('1.3.6.1.2.1.1.1.0'), DisplayString('Linux i386'))]])
    >>>

    """

    def __callback(
        snmpEngine: SnmpEngine,
        sendRequestHandle,
        errorIndication: errind.ErrorIndication,
        errorStatus: "Integer32 | int",
        errorIndex: "Integer32 | int",
        varBinds,
        cbCtx,
    ):
        lookupMib, future = cbCtx
        if future.cancelled():
            return

        if (
            options.get("ignoreNonIncreasingOid", False)
            and errorIndication
            and isinstance(errorIndication, errind.OidNotIncreasing)
        ):
            errorIndication = None  # TODO: fix this

        try:
            varBindsUnmade = VB_PROCESSOR.unmake_varbinds(
                snmpEngine.cache, varBinds, lookupMib
            )
        except Exception as e:
            future.set_exception(e)
        else:
            future.set_result((
                errorIndication,
                errorStatus,
                errorIndex,
                varBindsUnmade,
            ))

    addrName, paramsName = LCD.configure(
        snmpEngine, authData, transportTarget, contextData.contextName
    )

    future = asyncio.get_running_loop().create_future()

    cmdgen.NextCommandGenerator().send_varbinds(
        snmpEngine,
        addrName,
        contextData.contextEngineId,
        contextData.contextName,
        VB_PROCESSOR.make_varbinds(snmpEngine.cache, varBinds),
        __callback,
        (options.get("lookupMib", True), future),
    )
    return await future


async def bulk_cmd(
    snmpEngine: SnmpEngine,
    authData: "CommunityData | UsmUserData",
    transportTarget: AbstractTransportTarget,
    contextData: ContextData,
    nonRepeaters: int,
    maxRepetitions: int,
    *varBinds: ObjectType,
    **options,
) -> "tuple[errind.ErrorIndication, Integer32 | str | int, Integer32 | int, tuple[ObjectType, ...]]":
    r"""Creates a generator to perform SNMP GETBULK query.

    When iterator gets advanced by :py:mod:`asyncio` main loop,
    SNMP GETBULK request is send (:RFC:`1905#section-4.2.3`).
    The iterator yields :py:class:`asyncio.get_running_loop().create_future()` which gets done whenever
    response arrives or error occurs.

    Parameters
    ----------
    snmpEngine : :py:class:`~pysnmp.hlapi.v3arch.asyncio.SnmpEngine`
        Class instance representing SNMP engine.

    authData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.CommunityData` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.UsmUserData`
        Class instance representing SNMP credentials.

    transportTarget : :py:class:`~pysnmp.hlapi.v3arch.asyncio.UdpTransportTarget` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.Udp6TransportTarget`
        Class instance representing transport type along with SNMP peer address.

    contextData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.ContextData`
        Class instance representing SNMP ContextEngineId and ContextName values.

    nonRepeaters : int
        One MIB variable is requested in response for the first
        `nonRepeaters` MIB variables in request.

    maxRepetitions : int
        `maxRepetitions` MIB variables are requested in response for each
        of the remaining MIB variables in the request (e.g. excluding
        `nonRepeaters`). Remote SNMP engine may choose lesser value than
        requested.

    \*varBinds : :py:class:`~pysnmp.smi.rfc1902.ObjectType`
        One or more class instances representing MIB variables to place
        into SNMP request.

    Other Parameters
    ----------------
    \*\*options :
        Request options:

            * `lookupMib` - load MIB and resolve response MIB variables at
              the cost of slightly reduced performance. Default is `True`.
            * `ignoreNonIncreasingOid` - continue iteration even if response
              MIB variables (OIDs) are not greater then request MIB variables.
              Be aware that setting it to `True` may cause infinite loop between
              SNMP management and agent applications. Default is `False`.

    Yields
    ------
    errorIndication : :py:class:`~pysnmp.proto.errind.ErrorIndication`
        True value indicates SNMP engine error.
    errorStatus : str
        True value indicates SNMP PDU error.
    errorIndex : int
        Non-zero value refers to `varBinds[errorIndex-1]`
    varBindTable : tuple
        A sequence of sequences (e.g. 2-D array) of
        :py:class:`~pysnmp.smi.rfc1902.ObjectType` class instances
        representing a table of MIB variables returned in SNMP response, with
        up to ``maxRepetitions`` rows, i.e.
        ``len(varBindTable) <= maxRepetitions``.

        For ``0 <= i < len(varBindTable)`` and ``0 <= j < len(varBinds)``,
        ``varBindTable[i][j]`` represents:

        - For non-repeaters (``j < nonRepeaters``), the first lexicographic
          successor of ``varBinds[j]``, regardless the value of ``i``, or an
          :py:class:`~pysnmp.smi.rfc1902.ObjectType` instance with the
          :py:obj:`~pysnmp.proto.rfc1905.endOfMibView` value if no such
          successor exists;
        - For repeaters (``j >= nonRepeaters``), the ``i``-th lexicographic
          successor of ``varBinds[j]``, or an
          :py:class:`~pysnmp.smi.rfc1902.ObjectType` instance with the
          :py:obj:`~pysnmp.proto.rfc1905.endOfMibView` value if no such
          successor exists.

        See :rfc:`3416#section-4.2.3` for details on the underlying
        ``GetBulkRequest-PDU`` and the associated ``GetResponse-PDU``, such as
        specific conditions under which the server may truncate the response,
        causing ``varBindTable`` to have less than ``maxRepetitions`` rows.

    Raises
    ------
    PySnmpError
        Or its derivative indicating that an error occurred while
        performing SNMP operation.

    Examples
    --------
    >>> import asyncio
    >>> from pysnmp.hlapi.v3arch.asyncio import *
    >>>
    >>> async def run():
    ...     errorIndication, errorStatus, errorIndex, varBinds = await bulk_cmd(
    ...         SnmpEngine(),
    ...         CommunityData('public'),
    ...         await UdpTransportTarget.create(('demo.pysnmp.com', 161)),
    ...         ContextData(),
    ...         0, 2,
    ...         ObjectType(ObjectIdentity('SNMPv2-MIB', 'system'))
    ...     )
    ...     print(errorIndication, errorStatus, errorIndex, varBinds)
    >>>
    >>> asyncio.run(run())
    (None, 0, 0, [[ObjectType(ObjectIdentity(ObjectName('1.3.6.1.2.1.1.1.0')), DisplayString('SunOS zeus.pysnmp.com 4.1.3_U1 1 sun4m'))], [ObjectType(ObjectIdentity(ObjectName('1.3.6.1.2.1.1.2.0')), ObjectIdentifier('1.3.6.1.4.1.424242.1.1'))]])
    >>>

    """

    def __callback(
        snmpEngine: SnmpEngine,
        sendRequestHandle,
        errorIndication: errind.ErrorIndication,
        errorStatus: "Integer32 | int",
        errorIndex: "Integer32 | int",
        varBinds,
        cbCtx,
    ):
        lookupMib, future = cbCtx
        if future.cancelled():
            return

        if (
            options.get("ignoreNonIncreasingOid", False)
            and errorIndication
            and isinstance(errorIndication, errind.OidNotIncreasing)
        ):
            errorIndication = None  # TODO: fix here

        try:
            varBindsUnmade = VB_PROCESSOR.unmake_varbinds(
                snmpEngine.cache, varBinds, lookupMib
            )
        except Exception as e:
            future.set_exception(e)
        else:
            future.set_result((
                errorIndication,
                errorStatus,
                errorIndex,
                varBindsUnmade,
            ))

    addrName, paramsName = LCD.configure(
        snmpEngine, authData, transportTarget, contextData.contextName
    )

    future = asyncio.get_running_loop().create_future()

    cmdgen.BulkCommandGenerator().send_varbinds(
        snmpEngine,
        addrName,
        contextData.contextEngineId,
        contextData.contextName,
        nonRepeaters,
        maxRepetitions,
        VB_PROCESSOR.make_varbinds(snmpEngine.cache, varBinds),
        __callback,
        (options.get("lookupMib", True), future),
    )
    return await future


async def walk_cmd(
    snmpEngine: SnmpEngine,
    authData: "CommunityData | UsmUserData",
    transportTarget: AbstractTransportTarget,
    contextData: ContextData,
    varBind: ObjectType,
    **options,
) -> AsyncGenerator[
    "tuple[errind.ErrorIndication | None, Integer32 | str | int | None, Integer32 | int | None, tuple[ObjectType, ...]]",
    None,
]:
    r"""Creates a generator to perform one or more SNMP GETNEXT queries.

    On each iteration, new SNMP GETNEXT request is send
    (:RFC:`1905#section-4.2.2`). The iterator blocks waiting for response
    to arrive or error to occur.

    Parameters
    ----------
    snmpEngine : :py:class:`~pysnmp.hlapi.v3arch.asyncio.SnmpEngine`
        Class instance representing SNMP engine.

    authData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.CommunityData` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.UsmUserData`
        Class instance representing SNMP credentials.

    transportTarget : :py:class:`~pysnmp.hlapi.v3arch.asyncio.UdpTransportTarget` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.Udp6TransportTarget`
        Class instance representing transport type along with SNMP peer address.

    contextData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.ContextData`
        Class instance representing SNMP ContextEngineId and ContextName values.

    varBind : :py:class:`~pysnmp.smi.rfc1902.ObjectType`
        One class instance representing MIB variables to place
        into SNMP request.

    Other Parameters
    ----------------
    \*\*options :
        Request options:

            * `lookupMib` - load MIB and resolve response MIB variables at
              the cost of slightly reduced performance. Default is `True`.
            * `lexicographicMode` - walk SNMP agent's MIB till the end (if `True`),
              otherwise (if `False`) stop iteration when all response MIB
              variables leave the scope of initial MIB variables in
              `varBinds`. Default is `True`.
            * `ignoreNonIncreasingOid` - continue iteration even if response
              MIB variables (OIDs) are not greater then request MIB variables.
              Be aware that setting it to `True` may cause infinite loop between
              SNMP management and agent applications. Default is `False`.
            * `maxRows` - stop iteration once this generator instance processed
              `maxRows` of SNMP conceptual table. Default is `0` (no limit).
            * `maxCalls` - stop iteration once this generator instance processed
              `maxCalls` responses. Default is 0 (no limit).

    Yields
    ------
    errorIndication : str
        True value indicates SNMP engine error.
    errorStatus : str
        True value indicates SNMP PDU error.
    errorIndex : int
        Non-zero value refers to `varBinds[errorIndex-1]`
    varBinds : tuple
        A sequence of :py:class:`~pysnmp.smi.rfc1902.ObjectType` class
        instances representing MIB variables returned in SNMP response.

    Raises
    ------
    PySnmpError
        Or its derivative indicating that an error occurred while
        performing SNMP operation.

    Notes
    -----
    The `walk_cmd` generator will be exhausted on any of the following
    conditions:

    * SNMP engine error occurs thus `errorIndication` is `True`
    * SNMP PDU `errorStatus` is reported as `True`
    * SNMP :py:class:`~pysnmp.proto.rfc1905.EndOfMibView` values
      (also known as *SNMP exception values*) are reported for all
      MIB variables in `varBinds`
    * *lexicographicMode* option is `True` and SNMP agent reports
      end-of-mib or *lexicographicMode* is `False` and all
      response MIB variables leave the scope of `varBinds`

    At any moment a new sequence of `varBinds` could be send back into
    running generator (supported since Python 2.6).

    Examples
    --------
    >>> from pysnmp.hlapi.v3arch.asyncio import *
    >>> objects = walk_cmd(SnmpEngine(),
    ...             CommunityData('public'),
    ...             await UdpTransportTarget.create(('demo.pysnmp.com', 161)),
    ...             ContextData(),
    ...             ObjectType(ObjectIdentity('SNMPv2-MIB', 'sysDescr')))
    ... g = [item async for item in objects]
    >>> next(g)
    (None, 0, 0, [ObjectType(ObjectIdentity(ObjectName('1.3.6.1.2.1.1.1.0')), DisplayString('SunOS zeus.pysnmp.com 4.1.3_U1 1 sun4m'))])
    >>> g.send( [ ObjectType(ObjectIdentity('IF-MIB', 'ifInOctets')) ] )
    (None, 0, 0, [(ObjectName('1.3.6.1.2.1.2.2.1.10.1'), Counter32(284817787))])
    """
    lexicographicMode = options.get("lexicographicMode", True)
    ignoreNonIncreasingOid = options.get("ignoreNonIncreasingOid", False)
    maxRows = options.get("maxRows", 0)
    maxCalls = options.get("maxCalls", 0)

    initialVars = [
        x[0] for x in VB_PROCESSOR.make_varbinds(snmpEngine.cache, (varBind,))
    ]

    totalRows = totalCalls = 0

    while True:
        if varBind:
            errorIndication, errorStatus, errorIndex, varBindTable = await next_cmd(
                snmpEngine,
                authData,
                transportTarget,
                contextData,
                (varBind[0], Null("")),
                **dict(lookupMib=options.get("lookupMib", True)),
            )
            if (
                ignoreNonIncreasingOid
                and errorIndication
                and isinstance(errorIndication, errind.OidNotIncreasing)
            ):
                errorIndication = None

            if errorIndication:
                yield (errorIndication, errorStatus, errorIndex, (varBind,))
                return
            elif errorStatus:
                if errorStatus == 2:
                    # Hide SNMPv1 noSuchName error which leaks in here
                    # from SNMPv1 Agent through internal pysnmp proxy.
                    errorStatus = 0
                    errorIndex = 0
                return
            else:
                stopFlag = True

                varBind = varBindTable[0]

                name, val = varBind
                foundEnding = isinstance(val, Null) or isinstance(val, EndOfMibView)
                foundBeyond = not lexicographicMode and not initialVars[0].isPrefixOf(
                    name
                )
                if foundEnding or foundBeyond:
                    return

                if stopFlag and varBind[1] is not endOfMibView:
                    stopFlag = False

                if stopFlag:
                    return

                totalRows += 1
                totalCalls += 1
        else:
            errorIndication = errorStatus = errorIndex = None
            varBind = None

        initialVarBinds: "tuple[ObjectType, ...]|None" = yield (
            errorIndication,
            errorStatus,
            errorIndex,
            (varBind,),
        )

        if initialVarBinds:
            varBind = initialVarBinds[0]
            initialVars = [
                x[0]
                for x in VB_PROCESSOR.make_varbinds(snmpEngine.cache, initialVarBinds)
            ]

        if maxRows and totalRows >= maxRows:
            return

        if maxCalls and totalCalls >= maxCalls:
            return


async def bulk_walk_cmd(
    snmpEngine: SnmpEngine,
    authData: "CommunityData | UsmUserData",
    transportTarget: AbstractTransportTarget,
    contextData: ContextData,
    nonRepeaters: int,
    maxRepetitions: int,
    varBind: ObjectType,
    **options,
) -> AsyncGenerator[
    "tuple[errind.ErrorIndication | None, Integer32 | str | int | None, Integer32 | int | None, tuple[ObjectType, ...]]",
    None,
]:
    r"""Creates a generator to perform one or more SNMP GETBULK queries.

    On each iteration, new SNMP GETBULK request is send
    (:RFC:`1905#section-4.2.3`). The iterator blocks waiting for response
    to arrive or error to occur.

    Parameters
    ----------
    snmpEngine : :py:class:`~pysnmp.hlapi.v3arch.asyncio.SnmpEngine`
        Class instance representing SNMP engine.

    authData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.CommunityData` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.UsmUserData`
        Class instance representing SNMP credentials.

    transportTarget : :py:class:`~pysnmp.hlapi.v3arch.asyncio.UdpTransportTarget` or :py:class:`~pysnmp.hlapi.v3arch.asyncio.Udp6TransportTarget`
        Class instance representing transport type along with SNMP peer address.

    contextData : :py:class:`~pysnmp.hlapi.v3arch.asyncio.ContextData`
        Class instance representing SNMP ContextEngineId and ContextName values.

    nonRepeaters : int
        One MIB variable is requested in response for the first
        `nonRepeaters` MIB variables in request.

    maxRepetitions : int
        `maxRepetitions` MIB variables are requested in response for each
        of the remaining MIB variables in the request (e.g. excluding
        `nonRepeaters`). Remote SNMP engine may choose lesser value than
        requested.

    varBind : :py:class:`~pysnmp.smi.rfc1902.ObjectType`
        One class instance representing MIB variables to place
        into SNMP request.

    Other Parameters
    ----------------
    \*\*options :
        Request options:

            * `lookupMib` - load MIB and resolve response MIB variables at
              the cost of slightly reduced performance. Default is `True`.
            * `lexicographicMode` - walk SNMP agent's MIB till the end (if `True`),
              otherwise (if `False`) stop iteration when all response MIB
              variables leave the scope of initial MIB variables in
              `varBinds`. Default is `True`.
            * `ignoreNonIncreasingOid` - continue iteration even if response
              MIB variables (OIDs) are not greater then request MIB variables.
              Be aware that setting it to `True` may cause infinite loop between
              SNMP management and agent applications. Default is `False`.
            * `maxRows` - stop iteration once this generator instance processed
              `maxRows` of SNMP conceptual table. Default is `0` (no limit).
            * `maxCalls` - stop iteration once this generator instance processed
              `maxCalls` responses. Default is 0 (no limit).

    Yields
    ------
    errorIndication : str
        True value indicates SNMP engine error.
    errorStatus : str
        True value indicates SNMP PDU error.
    errorIndex : int
        Non-zero value refers to \*varBinds[errorIndex-1]
    varBinds : tuple
        A sequence of :py:class:`~pysnmp.smi.rfc1902.ObjectType` class
        instances representing MIB variables returned in SNMP response.

    Raises
    ------
    PySnmpError
        Or its derivative indicating that an error occurred while
        performing SNMP operation.

    Notes
    -----
    The `bulk_walk_cmd` generator will be exhausted on any of the following
    conditions:

    * SNMP engine error occurs thus `errorIndication` is `True`
    * SNMP PDU `errorStatus` is reported as `True`
    * SNMP :py:class:`~pysnmp.proto.rfc1905.EndOfMibView` values
      (also known as *SNMP exception values*) are reported for all
      MIB variables in `varBinds`
    * *lexicographicMode* option is `True` and SNMP agent reports
      end-of-mib or *lexicographicMode* is `False` and all
      response MIB variables leave the scope of `varBinds`

    At any moment a new sequence of `varBinds` could be send back into
    running generator (supported since Python 2.6).

    Setting `maxRepetitions` value to 15..50 might significantly improve
    system performance, as many MIB variables get packed into a single
    response message at once.

    Examples
    --------
    >>> from pysnmp.hlapi.v3arch.asyncio import *
    >>> objects = bulk_walk_cmd(SnmpEngine(),
    ...             CommunityData('public'),
    ...             await UdpTransportTarget.create(('demo.pysnmp.com', 161)),
    ...             ContextData(),
    ...             0, 25,
    ...             ObjectType(ObjectIdentity('SNMPv2-MIB', 'sysDescr')))
    ... g = [item async for item in objects]
    >>> next(g)
    (None, 0, 0, [ObjectType(ObjectIdentity(ObjectName('1.3.6.1.2.1.1.1.0')), DisplayString('SunOS zeus.pysnmp.com 4.1.3_U1 1 sun4m'))])
    >>> g.send( [ ObjectType(ObjectIdentity('IF-MIB', 'ifInOctets')) ] )
    (None, 0, 0, [(ObjectName('1.3.6.1.2.1.2.2.1.10.1'), Counter32(284817787))])
    """
    lexicographicMode = options.get("lexicographicMode", True)
    ignoreNonIncreasingOid = options.get("ignoreNonIncreasingOid", False)
    maxRows = options.get("maxRows", 0)
    maxCalls = options.get("maxCalls", 0)

    initialVars = [
        x[0] for x in VB_PROCESSOR.make_varbinds(snmpEngine.cache, (varBind,))
    ]

    totalRows = totalCalls = 0

    varBinds: "tuple[ObjectType, ...]" = (varBind,)

    while True:
        if maxRows and totalRows < maxRows:
            maxRepetitions = min(maxRepetitions, maxRows - totalRows)

        if varBinds:
            # Create a simple tuple with the OID from the previous response and Null value
            # This approach matches walk_cmd() and works with both lookupMib=True and False
            nextVarBinds = [(varBinds[-1][0], Null(""))]

            errorIndication, errorStatus, errorIndex, varBindTable = await bulk_cmd(
                snmpEngine,
                authData,
                transportTarget,
                contextData,
                nonRepeaters,
                maxRepetitions,
                *nextVarBinds,
                **dict(lookupMib=options.get("lookupMib", True)),
            )

            if (
                ignoreNonIncreasingOid
                and errorIndication
                and isinstance(errorIndication, errind.OidNotIncreasing)
            ):
                errorIndication = None

            if errorIndication:
                yield (
                    errorIndication,
                    errorStatus,
                    errorIndex,
                    varBindTable and varBinds,
                )
                if errorIndication != errind.requestTimedOut:
                    return
            elif errorStatus:
                if errorStatus == 2:
                    # Hide SNMPv1 noSuchName error which leaks in here
                    # from SNMPv1 Agent through internal pysnmp proxy.
                    errorStatus = 0
                    errorIndex = 0
                yield (
                    errorIndication,
                    errorStatus,
                    errorIndex,
                    varBinds,
                )
                return
            else:
                stopFlag = True
                varBinds = varBindTable

                for col, varBind in enumerate(varBinds):
                    name, val = varBind
                    foundEnding = isinstance(val, Null) or isinstance(val, EndOfMibView)
                    foundBeyond = not lexicographicMode and not initialVars[
                        0
                    ].isPrefixOf(name)
                    if foundEnding or foundBeyond:
                        stopFlag = True
                        result = varBinds[:col]
                        if len(result) > 0:
                            yield (
                                errorIndication,
                                errorStatus,
                                errorIndex,
                                result,
                            )
                        return

                    if stopFlag and varBinds[col][1] is not endOfMibView:
                        stopFlag = False

                if stopFlag:
                    return

                totalRows += 1
                totalCalls += 1
        else:
            errorIndication = errorStatus = errorIndex = None
            varBinds = ()

        initialVarBinds: "tuple[ObjectType, ...]|None" = yield (
            errorIndication,
            errorStatus,
            errorIndex,
            varBinds,
        )
        if initialVarBinds:
            varBinds = initialVarBinds
            initialVars = [
                x[0] for x in VB_PROCESSOR.make_varbinds(snmpEngine.cache, varBinds)
            ]

        if maxRows and totalRows >= maxRows:
            return

        if maxCalls and totalCalls >= maxCalls:
            return


# Compatibility API
deprecated_attributes = {
    "getCmd": "get_cmd",
    "setCmd": "set_cmd",
    "nextCmd": "next_cmd",
    "bulkCmd": "bulk_cmd",
    "walkCmd": "walk_cmd",
    "bulkWalkCmd": "bulk_walk_cmd",
    "isEndOfMib": "is_end_of_mib",
}


def __getattr__(attr: str):
    if new_attr := deprecated_attributes.get(attr):
        warnings.warn(
            f"{attr} is deprecated. Please use {new_attr} instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return globals()[new_attr]
    raise AttributeError(f"module '{__name__}' has no attribute '{attr}'")



#!/usr/bin/env python3
"""
Script to run tests for the project.
Equivalent to Run-Tests.ps1
"""

import argparse
import os
import platform
import subprocess
import sys
from pathlib import Path


def main():
    """Main function to run tests."""
    # Parse command line arguments
    parser = argparse.ArgumentParser(description="Run tests for this project.")
    parser.add_argument(
        "--coverage", action="store_true", help="Run with coverage report"
    )
    parser.add_argument(
        "--verbose", action="store_true", help="Run with verbose output"
    )
    parser.add_argument("--test-file", default="", help="Specific test file to run")
    parser.add_argument("--test-path", default="tests", help="Path to test directory")
    parser.add_argument("--args", default="", help="Additional pytest arguments")

    args = parser.parse_args()

    # Check if virtual environment is active
    if not Path(".venv").exists():
        print("No virtual environment found. Creating one...", file=sys.stderr)
        subprocess.run(["uv", "venv"], check=True)

        # Activate the virtual environment
        # Note: Python doesn't need to explicitly "activate" the venv like PowerShell does
        # We can use the Python from the venv directly

        print("Installing dependencies...", file=sys.stderr)
        subprocess.run(["uv", "pip", "install", "-e", ".[dev]"], check=True)

    print("Running tests for this project...", file=sys.stderr)

    # Build the command
    command = ["uv", "run", "pytest"]

    if args.verbose:
        command.append("-v")

    if args.coverage:
        command.extend(["--cov=pysnmp", "--cov-report=term", "--cov-report=html"])

    if args.test_file:
        command.append(args.test_file)
    else:
        command.append(args.test_path)

    if args.args:
        command.extend(args.args.split())

    # Display the command being run
    print(f"Executing: {' '.join(command)}", file=sys.stderr)

    # Run the command
    subprocess.run(command, check=True)

    # Open coverage report if generated
    if args.coverage:
        print("\nCoverage report generated in htmlcov/index.html", file=sys.stderr)

        if platform.system() == "Windows":
            os.startfile("htmlcov\\index.html")
        elif platform.system() == "Darwin":  # macOS
            subprocess.run(["open", "htmlcov/index.html"], check=True)
        elif platform.system() == "Linux":
            subprocess.run(["xdg-open", "htmlcov/index.html"], check=True)


if __name__ == "__main__":
    main()



import os
import pytest
from pysnmp.smi import builder, compiler, view

mibBuilder = builder.MibBuilder()
mibBuilder.loadTexts = True
mibBuilder.add_mib_sources(builder.DirMibSource(compiler.DEFAULT_DEST))
mibView = view.MibViewController(mibBuilder)

mibBuilder.load_modules(
    "SNMPv2-MIB",
    "SNMP-FRAMEWORK-MIB",
    "SNMP-COMMUNITY-MIB",
    "CISCO-ENHANCED-IPSEC-FLOW-MIB",
    "PYSNMP-MIB",
)


def test_getNodeName_by_OID():
    oid, label, suffix = mibView.get_node_name((1, 3, 6, 1, 2, 1, 1, 1))
    assert oid == (1, 3, 6, 1, 2, 1, 1, 1)
    assert label == (
        "iso",
        "org",
        "dod",
        "internet",
        "mgmt",
        "mib-2",
        "system",
        "sysDescr",
    )
    assert suffix == ()


def test_getNodeName_by_label():
    oid, label, suffix = mibView.get_node_name((1, 3, 6, 1, 2, "mib-2", 1, "sysDescr"))
    assert oid == (1, 3, 6, 1, 2, 1, 1, 1)
    assert label == (
        "iso",
        "org",
        "dod",
        "internet",
        "mgmt",
        "mib-2",
        "system",
        "sysDescr",
    )
    assert suffix == ()


def test_getNodeName_by_symbol_description():
    oid, label, suffix = mibView.get_node_name(("sysDescr",))
    assert oid == (1, 3, 6, 1, 2, 1, 1, 1)
    assert label == (
        "iso",
        "org",
        "dod",
        "internet",
        "mgmt",
        "mib-2",
        "system",
        "sysDescr",
    )
    assert suffix == ()


def test_getNodeName_by_symbol_description_with_module_name():
    oid, label, suffix = mibView.get_node_name(("snmpEngineID",), "SNMP-FRAMEWORK-MIB")
    assert oid == (1, 3, 6, 1, 6, 3, 10, 2, 1, 1)
    assert label == (
        "iso",
        "org",
        "dod",
        "internet",
        "snmpV2",
        "snmpModules",
        "snmpFrameworkMIB",
        "snmpFrameworkMIBObjects",
        "snmpEngine",
        "snmpEngineID",
    )
    assert suffix == ()

    (mibNode,) = mibBuilder.import_symbols("SNMP-FRAMEWORK-MIB", "snmpEngineID")
    assert "" != mibNode.syntax.prettyPrint()

    # Method 1: Assert that mibNode.syntax is an instance of OctetString
    from pysnmp.proto.rfc1902 import OctetString

    assert isinstance(mibNode.syntax, OctetString), (
        "mibNode.syntax is not an instance of OctetString"
    )

    # Method 2: When you don't know the base class in advance
    # Extract and print the class hierarchy of mibNode.syntax
    syntax_type = type(mibNode.syntax)
    class_hierarchy = []
    for base in syntax_type.__mro__:
        class_hierarchy.append(base.__name__)

    # Check if 'OctetString' is in the class hierarchy
    assert "OctetString" in class_hierarchy, (
        f"OctetString not found in class hierarchy: {class_hierarchy}"
    )

    # Print the class hierarchy for reference
    print(f"Class hierarchy of mibNode.syntax: {class_hierarchy}")

    # Method 3: Find the immediate subclass of SimpleAsn1Type
    from pyasn1.type.base import SimpleAsn1Type

    # Get the full MRO as classes, not just names
    full_mro = syntax_type.__mro__

    # Find SimpleAsn1Type in the MRO
    simple_asn_index = None
    for i, cls in enumerate(full_mro):
        if cls is SimpleAsn1Type:
            simple_asn_index = i
            break

    # If found, the direct subclass is at index - 1, and the subclass above that is at index - 2
    assert simple_asn_index is not None, "SimpleAsn1Type not found in class hierarchy"

    # The class at index - 1 is the immediate subclass of SimpleAsn1Type
    # The class at index - 2 is the subclass above the immediate subclass (i.e., the essential SMI/SNMP data type)
    essential_type = full_mro[simple_asn_index - 2]
    print(f"Essential SMI/SNMP data type: {essential_type.__name__}")

    # Assert that the essential type is OctetString
    assert essential_type is OctetString, (
        f"Essential type is {essential_type.__name__}, not OctetString"
    )


def test_getUnits_by_symbol_description_with_module_name():
    oid, label, suffix = mibView.get_node_name(
        ("snmpEngineTime",), "SNMP-FRAMEWORK-MIB"
    )
    assert oid == (1, 3, 6, 1, 6, 3, 10, 2, 1, 3)
    assert label == (
        "iso",
        "org",
        "dod",
        "internet",
        "snmpV2",
        "snmpModules",
        "snmpFrameworkMIB",
        "snmpFrameworkMIBObjects",
        "snmpEngine",
        "snmpEngineTime",
    )
    assert suffix == ()

    (mibNode,) = mibBuilder.import_symbols("SNMP-FRAMEWORK-MIB", "snmpEngineTime")
    # assert "" != mibNode.syntax.prettyPrint()

    assert mibNode.getUnits() == "seconds"


def test_getNodeName_by_symbol_location_lookup_by_name():
    modName, symName, suffix = mibView.get_node_location(("snmpCommunityEntry",))
    assert modName == "SNMP-COMMUNITY-MIB"
    assert symName == "snmpCommunityEntry"
    assert suffix == ()


def test_getNodeName_by_symbol_description_with_module_name_2():
    oid, label, suffix = mibView.get_node_name(
        ("ciscoEnhIPsecFlowActivityGroup",), "CISCO-ENHANCED-IPSEC-FLOW-MIB"
    )
    assert oid == (1, 3, 6, 1, 4, 1, 9, 9, 432, 2, 2, 1)
    assert label == (
        "iso",
        "org",
        "dod",
        "internet",
        "private",
        "enterprises",
        "cisco",
        "ciscoMgmt",
        "ciscoEnhancedIpsecFlowMIB",
        "ciscoEnhancedIpsecFlowMIBConform",
        "ciscoIPsecFlowMIBGroups",
        "ciscoEnhIPsecFlowActivityGroup",
    )
    assert suffix == ()

    (mibNode,) = mibBuilder.import_symbols(
        "CISCO-ENHANCED-IPSEC-FLOW-MIB", "ciscoEnhIPsecFlowActivityGroup"
    )
    assert (
        "This group consists of: 1) IPsec Phase-2 Global Statistics 2) IPsec Phase-2 Tunnel Table 3) IPsec Phase-2 Endpoint Table 4) IPsec Phase-2 Security Association Table"
        == mibNode.getDescription()
    )
    assert "rfc2408, rfc2407; rfc2409 section 5.5" == mibNode.getReference()



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from pysnmp.proto import errind, error
from pysnmp.proto.secmod.rfc3414.priv import base


class NoPriv(base.AbstractEncryptionService):
    """No privacy protocol.

    This class implements no privacy protocol.
    """

    SERVICE_ID = (1, 3, 6, 1, 6, 3, 10, 1, 2, 1)  # usmNoPrivProtocol

    def hash_passphrase(self, authProtocol, privKey):
        """Hash a passphrase."""
        return

    def localize_key(self, authProtocol, privKey, snmpEngineID):
        """Localize key."""
        return

    def encrypt_data(self, encryptKey, privParameters, dataToEncrypt):
        """Encrypt data."""
        raise error.StatusInformation(errorIndication=errind.noEncryption)

    def decrypt_data(self, decryptKey, privParameters, encryptedData):
        """Decrypt data."""
        raise error.StatusInformation(errorIndication=errind.noEncryption)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# PySNMP MIB module SNMPv2-CONF (https://www.pysnmp.com/pysnmp)
#

(MibNode,) = mibBuilder.import_symbols("SNMPv2-SMI", "MibNode")


class ObjectGroup(MibNode):
    status = "current"
    objects = ()
    description = ""
    reference = ""

    def getStatus(self):
        return self.status

    def setStatus(self, v):
        self.status = v
        return self

    def getObjects(self):
        return getattr(self, "objects", ())

    def setObjects(self, *args, **kwargs):
        if kwargs.get("append"):
            self.objects += args
        else:
            self.objects = args
        return self

    def getDescription(self):
        return getattr(self, "description", "")

    def setDescription(self, v):
        self.description = v
        return self

    def getReference(self):
        return self.reference

    def setReference(self, v):
        self.reference = v
        return self

    def asn1Print(self):
        return """\
OBJECT-GROUP
  OBJECTS {{ {} }}
  STATUS "{}"
  DESCRIPTION "{}"
""".format(
            ", ".join([x for x in self.getObjects()]),
            self.status,
            self.getDescription(),
        )

    ## compatibility with legacy code
    def set_objects(self, *args, **kwargs):
        return self.setObjects(*args, **kwargs)

    def set_description(self, v):
        return self.setDescription(v)

    def set_status(self, v):
        return self.setStatus(v)

    def set_reference(self, v):
        return self.setReference(v)


class NotificationGroup(MibNode):
    status = "current"
    objects = ()
    description = ""
    reference = ""

    def getStatus(self):
        return self.status

    def setStatus(self, v):
        self.status = v
        return self

    def getObjects(self):
        return getattr(self, "objects", ())

    def setObjects(self, *args, **kwargs):
        if kwargs.get("append"):
            self.objects += args
        else:
            self.objects = args
        return self

    def getDescription(self):
        return getattr(self, "description", "")

    def setDescription(self, v):
        self.description = v
        return self

    def getReference(self):
        return self.reference

    def setReference(self, v):
        self.reference = v
        return self

    def asn1Print(self):
        return """\
NOTIFICATION-GROUP
  NOTIFICATIONS {{ {} }}
  STATUS "{}"
  DESCRIPTION "{}"
""".format(
            ", ".join([x for x in self.getObjects()]),
            self.getStatus(),
            self.getDescription(),
        )

    ## compatibility with legacy code
    def set_objects(self, *args, **kwargs):
        return self.setObjects(*args, **kwargs)

    def set_description(self, v):
        return self.setDescription(v)

    def set_status(self, v):
        return self.setStatus(v)

    def set_reference(self, v):
        return self.setReference(v)


class ModuleCompliance(MibNode):
    status = "current"
    objects = ()
    description = ""

    def getStatus(self):
        return self.status

    def setStatus(self, v):
        self.status = v
        return self

    def getObjects(self):
        return getattr(self, "objects", ())

    def setObjects(self, *args, **kwargs):
        if kwargs.get("append"):
            self.objects += args
        else:
            self.objects = args
        return self

    def getDescription(self):
        return getattr(self, "description", "")

    def setDescription(self, v):
        self.description = v
        return self

    def getReference(self):
        return self.reference

    def setReference(self, v):
        self.reference = v
        return self

    def asn1Print(self):
        return """\
MODULE-COMPLIANCE
  STATUS "{}"
  DESCRIPTION "{}"
  OBJECT {{ {} }}
""".format(
            self.getStatus(),
            self.getDescription(),
            ", ".join([x for x in self.getObjects()]),
        )

    ## compatibility with legacy code
    def set_objects(self, *args, **kwargs):
        return self.setObjects(*args, **kwargs)

    def set_status(self, v):
        return self.setStatus(v)

    def set_description(self, v):
        return self.setDescription(v)


class AgentCapabilities(MibNode):
    status = "current"
    description = ""
    reference = ""
    productRelease = ""

    def getStatus(self):
        return self.status

    def setStatus(self, v):
        self.status = v
        return self

    def getDescription(self):
        return getattr(self, "description", "")

    def setDescription(self, v):
        self.description = v
        return self

    def getReference(self):
        return self.reference

    def setReference(self, v):
        self.reference = v
        return self

    def getProductRelease(self):
        return self.productRelease

    def setProductRelease(self, v):
        self.productRelease = v
        return self

    # TODO: implement the rest of properties

    def asn1Print(self):
        return """\
AGENT-CAPABILITIES
  PRODUCT-RELEASE "{}"
  STATUS "{}"
  DESCRIPTION "{}"
""".format(
            self.getProductRelease(), self.getStatus(), self.getDescription()
        )


mibBuilder.export_symbols(
    "SNMPv2-CONF",
    ObjectGroup=ObjectGroup,
    NotificationGroup=NotificationGroup,
    ModuleCompliance=ModuleCompliance,
    AgentCapabilities=AgentCapabilities,
)



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
"""
This module provides debugging and logging utilities for the pysnmp library.

Classes:
    Printer: A class for controlling logging output.
    Debug: A helper class for managing debug flags and logging messages.

Functions:
    __getattr__(attr: str): Handles deprecated attribute access and raises AttributeError for unknown attributes.
    setLogger(value): Sets the global logger.
    hexdump(octets): Returns a hexdump of the given octets.

Constants:
    FLAG_NONE: No flags set.
    FLAG_IO: Input/Output flag.
    FLAG_DSP: Dispatcher flag.
    FLAG_MP: Message Processing flag.
    FLAG_SM: Security Model flag.
    FLAG_BLD: MIB Builder flag.
    FLAG_MIB: MIB flag.
    FLAG_INS: Instrumentation flag.
    FLAG_ACL: Access Control List flag.
    FLAG_PRX: Proxy flag.
    FLAG_APP: Application flag.
    FLAG_ALL: All flags set.
    FLAG_MAP: A dictionary mapping flag names to their corresponding constants.

Attributes:
    NullHandler: A logging handler that does nothing.
    logger: A global logger variable.
"""
import logging
import warnings


from pysnmp import __version__, error

# Compatibility API
deprecated_attributes = {
    "flagMap": "FLAG_MAP",
    "flagNone": "FLAG_NONE",
    "flagIO": "FLAG_IO",
    "flagDSP": "FLAG_DSP",
    "flagMP": "FLAG_MP",
    "flagSM": "FLAG_SM",
    "flagBLD": "FLAG_BLD",
    "flagMIB": "FLAG_MIB",
    "flagINS": "FLAG_INS",
    "flagACL": "FLAG_ACL",
    "flagPRX": "FLAG_PRX",
    "flagAPP": "FLAG_APP",
    "flagALL": "FLAG_ALL",
}


def __getattr__(attr: str):
    """Handle deprecated attributes."""
    if newAttr := deprecated_attributes.get(attr):
        warnings.warn(
            f"{attr} is deprecated. Please use {newAttr} instead.", DeprecationWarning
        )
        return globals()[newAttr]
    raise AttributeError(f"module {__name__} has no attribute {attr}")


FLAG_NONE = 0x0000
FLAG_IO = 0x0001
FLAG_DSP = 0x0002
FLAG_MP = 0x0004
FLAG_SM = 0x0008
FLAG_BLD = 0x0010
FLAG_MIB = 0x0020
FLAG_INS = 0x0040
FLAG_ACL = 0x0080
FLAG_PRX = 0x0100
FLAG_APP = 0x0200
FLAG_ALL = 0xFFFF

FLAG_MAP = {
    "io": FLAG_IO,
    "dsp": FLAG_DSP,
    "msgproc": FLAG_MP,
    "secmod": FLAG_SM,
    "mibbuild": FLAG_BLD,
    "mibview": FLAG_MIB,
    "mibinstrum": FLAG_INS,
    "acl": FLAG_ACL,
    "proxy": FLAG_PRX,
    "app": FLAG_APP,
    "all": FLAG_ALL,
}


class Printer:
    """Logging control."""

    def __init__(self, logger=None, handler=None, formatter=None):
        """Logging control."""
        if logger is None:
            logger = logging.getLogger("pysnmp")
        logger.setLevel(logging.DEBUG)
        if handler is None:
            handler = logging.StreamHandler()
        if formatter is None:
            formatter = logging.Formatter("%(asctime)s %(name)s: %(message)s")
        handler.setFormatter(formatter)
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)
        self.__logger = logger

    def __call__(self, msg):
        """Log a message."""
        self.__logger.debug(msg)

    def __str__(self):
        """Return a string representation of the object."""
        return "<python built-in logging>"


NullHandler = logging.NullHandler


class Debug:
    """Logging control."""

    DEFAULT_PRINTER = None

    def __init__(self, *flags, **options):
        """Helper class for logging control."""
        self._flags = FLAG_NONE
        if options.get("printer") is not None:
            self._printer = options.get("printer")
        elif self.DEFAULT_PRINTER is not None:
            self._printer = self.DEFAULT_PRINTER
        else:
            if "loggerName" in options:
                # route our logs to parent logger
                self._printer = Printer(
                    logger=logging.getLogger(options["loggerName"]),
                    handler=NullHandler(),
                )
            else:
                self._printer = Printer()
        self("running pysnmp version %s" % __version__)
        for f in flags:
            inverse = f and f[0] in ("!", "~")
            if inverse:
                f = f[1:]
            try:
                if inverse:
                    self._flags &= ~FLAG_MAP[f]
                else:
                    self._flags |= FLAG_MAP[f]
            except KeyError:
                raise error.PySnmpError("bad debug flag %s" % f)

            self(
                "debug category '{}' {}".format(f, inverse and "disabled" or "enabled")
            )

    def __str__(self):
        """Return a string representation of the object."""
        return f"logger {self._printer}, flags {self._flags:x}"

    def __call__(self, msg):
        """Log a message."""
        self._printer(msg)

    def __and__(self, flag):
        """Return a boolean value of the flag."""
        return self._flags & flag

    def __rand__(self, flag):
        """Return a boolean value of the flag."""
        return flag & self._flags


# This will yield false from bitwise and with a flag, and save
# on unnecessary calls
logger = 0


def set_logger(value):
    """Set the global logger."""
    global logger
    logger = value


def hexdump(octets):
    """Return a hexdump of the given octets."""
    return " ".join(
        [
            "{}{:02X}".format(n % 16 == 0 and ("\n%.5d: " % n) or "", x)
            for n, x in zip(range(len(octets)), octets)
        ]
    )



"""
This module is part of the pysnmp library and is necessary to make the 'rfc7860.auth' directory a package.

Filepath:
    /Users/lextm/pysnmp.com/pysnmp/pysnmp/proto/secmod/rfc7860/auth/__init__.py
"""



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
from typing import Any, Dict, Tuple

from pysnmp.proto.api import v2c
from pysnmp.smi import builder, view
from pysnmp.smi.rfc1902 import NotificationType, ObjectIdentity, ObjectType

__all__ = ["CommandGeneratorVarBinds", "NotificationOriginatorVarBinds"]


def is_end_of_mib(var_binds):  # noqa: N816
    """
    Check if the given variable bindings indicate the end of the MIB.

    Parameters:
    var_binds (list): A list of variable bindings.

    Returns:
    bool: True if it is the end of the MIB, False otherwise.
    """
    return not v2c.apiPDU.get_next_varbinds(var_binds)[1]


class MibViewControllerManager:
    @staticmethod
    def get_mib_view_controller(userCache):
        try:
            mibViewController = userCache["mibViewController"]

        except KeyError:
            mibViewController = view.MibViewController(builder.MibBuilder())
            userCache["mibViewController"] = mibViewController

        return mibViewController


class CommandGeneratorVarBinds(MibViewControllerManager):
    """Var-binds processor for Command Generator."""

    def make_varbinds(
        self, userCache: Dict[str, Any], varBinds: Tuple[ObjectType, ...]
    ) -> Tuple[ObjectType, ...]:
        """Return a tuple of ObjectType instances."""
        mibViewController = self.get_mib_view_controller(userCache)

        resolvedVarBinds = []

        for varBind in varBinds:
            if isinstance(varBind, ObjectType):
                pass

            elif isinstance(varBind[0], ObjectIdentity):
                varBind = ObjectType(*varBind)

            elif isinstance(varBind[0][0], tuple):  # legacy
                varBind = ObjectType(
                    ObjectIdentity(varBind[0][0][0], varBind[0][0][1], *varBind[0][1:]),
                    varBind[1],
                )

            else:
                varBind = ObjectType(ObjectIdentity(varBind[0]), varBind[1])

            resolvedVarBinds.append(
                varBind.resolve_with_mib(mibViewController, ignoreErrors=False)
            )

        return tuple(resolvedVarBinds)

    def unmake_varbinds(
        self,
        userCache: Dict[str, Any],
        varBinds: Tuple[ObjectType, ...],
        lookupMib=True,
    ) -> Tuple[ObjectType, ...]:
        """Return a tuple of ObjectType instances."""
        if lookupMib:
            mibViewController = self.get_mib_view_controller(userCache)
            varBinds = tuple(
                ObjectType(ObjectIdentity(x[0]), x[1]).resolve_with_mib(
                    mibViewController
                )
                for x in varBinds
            )

        return varBinds


class NotificationOriginatorVarBinds(MibViewControllerManager):
    """Var-binds processor for Notification Originator."""

    def make_varbinds(
        self, userCache: Dict[str, Any], varBinds: "tuple[NotificationType, ...]"
    ) -> "tuple[ObjectType, ...]":
        """Return a tuple of ObjectType instances."""
        mibViewController = self.get_mib_view_controller(userCache)

        # TODO: this shouldn't be needed
        if isinstance(varBinds, NotificationType):
            return varBinds.resolve_with_mib(
                mibViewController, ignoreErrors=False
            ).to_varbinds()

        resolvedVarBinds: "list[ObjectType]" = []

        for varBind in varBinds:
            if isinstance(varBind, NotificationType):
                resolvedVarBinds.extend(
                    varBind.resolve_with_mib(
                        mibViewController, ignoreErrors=False
                    ).to_varbinds()
                )
                continue

            if isinstance(varBind, ObjectType):
                pass

            elif isinstance(varBind[0], ObjectIdentity):
                varBind = ObjectType(*varBind)

            else:
                varBind = ObjectType(ObjectIdentity(varBind[0]), varBind[1])

            resolvedVarBinds.append(
                varBind.resolve_with_mib(mibViewController, ignoreErrors=False)
            )

        return tuple(resolvedVarBinds)

    def unmake_varbinds(
        self,
        userCache: Dict[str, Any],
        varBinds: "tuple[ObjectType, ...]",
        lookupMib=False,
    ) -> "tuple[ObjectType, ...]":
        """Return a tuple of ObjectType instances."""
        if lookupMib:
            mibViewController = self.get_mib_view_controller(userCache)
            varBinds = tuple(
                ObjectType(ObjectIdentity(x[0]), x[1]).resolve_with_mib(
                    mibViewController
                )
                for x in varBinds
            )
        return varBinds



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# PySNMP MIB module INET-ADDRESS-MIB (https://www.pysnmp.com/pysnmp)
# ASN.1 source file://asn1/INET-ADDRESS-MIB
# Produced by pysmi-1.5.8 at Sat Nov  2 15:25:22 2024
# On host MacBook-Pro.local platform Darwin version 24.1.0 by user lextm
# Using Python version 3.12.0 (main, Nov 14 2023, 23:52:11) [Clang 15.0.0 (clang-1500.0.40.1)]
#
# IMPORTANT: customization

# Import base ASN.1 objects even if this MIB does not use it

(Integer,
 OctetString,
 ObjectIdentifier) = mibBuilder.import_symbols(
    "ASN1",
    "Integer",
    "OctetString",
    "ObjectIdentifier")

(NamedValues,) = mibBuilder.import_symbols(
    "ASN1-ENUMERATION",
    "NamedValues")
(ConstraintsIntersection,
 ConstraintsUnion,
 SingleValueConstraint,
 ValueRangeConstraint,
 ValueSizeConstraint) = mibBuilder.import_symbols(
    "ASN1-REFINEMENT",
    "ConstraintsIntersection",
    "ConstraintsUnion",
    "SingleValueConstraint",
    "ValueRangeConstraint",
    "ValueSizeConstraint")

# Import SMI symbols from the MIBs this MIB depends on

(ModuleCompliance,
 NotificationGroup) = mibBuilder.import_symbols(
    "SNMPv2-CONF",
    "ModuleCompliance",
    "NotificationGroup")

(Bits,
 Counter32,
 Counter64,
 Gauge32,
 Integer32,
 IpAddress,
 ModuleIdentity,
 MibIdentifier,
 NotificationType,
 ObjectIdentity,
 MibScalar,
 MibTable,
 MibTableRow,
 MibTableColumn,
 TimeTicks,
 Unsigned32,
 iso,
 mib_2) = mibBuilder.import_symbols(
    "SNMPv2-SMI",
    "Bits",
    "Counter32",
    "Counter64",
    "Gauge32",
    "Integer32",
    "IpAddress",
    "ModuleIdentity",
    "MibIdentifier",
    "NotificationType",
    "ObjectIdentity",
    "MibScalar",
    "MibTable",
    "MibTableRow",
    "MibTableColumn",
    "TimeTicks",
    "Unsigned32",
    "iso",
    "mib-2")

(DisplayString,
 TextualConvention) = mibBuilder.import_symbols(
    "SNMPv2-TC",
    "DisplayString",
    "TextualConvention")


# MODULE-IDENTITY

inetAddressMIB = ModuleIdentity(
    (1, 3, 6, 1, 2, 1, 76)
)
if mibBuilder.loadTexts:
    inetAddressMIB.setRevisions(
        ("2005-02-04 00:00",
         "2002-05-09 00:00",
         "2000-06-08 00:00")
    )
if mibBuilder.loadTexts:
    inetAddressMIB.setLastUpdated("200502040000Z")
if mibBuilder.loadTexts:
    inetAddressMIB.setOrganization("IETF Operations and Management Area")
if mibBuilder.loadTexts:
    inetAddressMIB.setContactInfo("Juergen Schoenwaelder (Editor) International University Bremen P.O. Box 750 561 28725 Bremen, Germany Phone: +49 421 200-3587 EMail: j.schoenwaelder@iu-bremen.de Send comments to <ietfmibs@ops.ietf.org>.")
if mibBuilder.loadTexts:
    inetAddressMIB.setDescription("This MIB module defines textual conventions for representing Internet addresses. An Internet address can be an IPv4 address, an IPv6 address, or a DNS domain name. This module also defines textual conventions for Internet port numbers, autonomous system numbers, and the length of an Internet address prefix. Copyright (C) The Internet Society (2005). This version of this MIB module is part of RFC 4001, see the RFC itself for full legal notices.")


# Types definitions


# TEXTUAL-CONVENTIONS



class InetAddressType(TextualConvention, Integer32):
    status = "current"
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1,
              2,
              3,
              4,
              16)
        )
    )
    namedValues = NamedValues(
        *(("unknown", 0),
          ("ipv4", 1),
          ("ipv6", 2),
          ("ipv4z", 3),
          ("ipv6z", 4),
          ("dns", 16))
    )

    if mibBuilder.loadTexts:
        description = "A value that represents a type of Internet address. unknown(0) An unknown address type. This value MUST be used if the value of the corresponding InetAddress object is a zero-length string. It may also be used to indicate an IP address that is not in one of the formats defined below. ipv4(1) An IPv4 address as defined by the InetAddressIPv4 textual convention. ipv6(2) An IPv6 address as defined by the InetAddressIPv6 textual convention. ipv4z(3) A non-global IPv4 address including a zone index as defined by the InetAddressIPv4z textual convention. ipv6z(4) A non-global IPv6 address including a zone index as defined by the InetAddressIPv6z textual convention. dns(16) A DNS domain name as defined by the InetAddressDNS textual convention. Each definition of a concrete InetAddressType value must be accompanied by a definition of a textual convention for use with that InetAddressType. To support future extensions, the InetAddressType textual convention SHOULD NOT be sub-typed in object type definitions. It MAY be sub-typed in compliance statements in order to require only a subset of these address types for a compliant implementation. Implementations must ensure that InetAddressType objects and any dependent objects (e.g., InetAddress objects) are consistent. An inconsistentValue error must be generated if an attempt to change an InetAddressType object would, for example, lead to an undefined InetAddress value. In particular, InetAddressType/InetAddress pairs must be changed together if the address type changes (e.g., from ipv6(2) to ipv4(1))."


class InetAddressIPv4(TextualConvention, OctetString):
    status = "current"
    displayHint = "1d.1d.1d.1d"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(4, 4),
    )
    fixed_length = 4

    if mibBuilder.loadTexts:
        description = "Represents an IPv4 network address: Octets Contents Encoding 1-4 IPv4 address network-byte order The corresponding InetAddressType value is ipv4(1). This textual convention SHOULD NOT be used directly in object definitions, as it restricts addresses to a specific format. However, if it is used, it MAY be used either on its own or in conjunction with InetAddressType, as a pair."


class InetAddressIPv6(TextualConvention, OctetString):
    status = "current"
    displayHint = "2x:2x:2x:2x:2x:2x:2x:2x"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(16, 16),
    )
    fixed_length = 16

    if mibBuilder.loadTexts:
        description = "Represents an IPv6 network address: Octets Contents Encoding 1-16 IPv6 address network-byte order The corresponding InetAddressType value is ipv6(2). This textual convention SHOULD NOT be used directly in object definitions, as it restricts addresses to a specific format. However, if it is used, it MAY be used either on its own or in conjunction with InetAddressType, as a pair."


class InetAddressIPv4z(TextualConvention, OctetString):
    status = "current"
    displayHint = "1d.1d.1d.1d%4d"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(8, 8),
    )
    fixed_length = 8

    if mibBuilder.loadTexts:
        description = "Represents a non-global IPv4 network address, together with its zone index: Octets Contents Encoding 1-4 IPv4 address network-byte order 5-8 zone index network-byte order The corresponding InetAddressType value is ipv4z(3). The zone index (bytes 5-8) is used to disambiguate identical address values on nodes that have interfaces attached to different zones of the same scope. The zone index may contain the special value 0, which refers to the default zone for each scope. This textual convention SHOULD NOT be used directly in object definitions, as it restricts addresses to a specific format. However, if it is used, it MAY be used either on its own or in conjunction with InetAddressType, as a pair."


class InetAddressIPv6z(TextualConvention, OctetString):
    status = "current"
    displayHint = "2x:2x:2x:2x:2x:2x:2x:2x%4d"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(20, 20),
    )
    fixed_length = 20

    if mibBuilder.loadTexts:
        description = "Represents a non-global IPv6 network address, together with its zone index: Octets Contents Encoding 1-16 IPv6 address network-byte order 17-20 zone index network-byte order The corresponding InetAddressType value is ipv6z(4). The zone index (bytes 17-20) is used to disambiguate identical address values on nodes that have interfaces attached to different zones of the same scope. The zone index may contain the special value 0, which refers to the default zone for each scope. This textual convention SHOULD NOT be used directly in object definitions, as it restricts addresses to a specific format. However, if it is used, it MAY be used either on its own or in conjunction with InetAddressType, as a pair."


class InetAddressDNS(TextualConvention, OctetString):
    status = "current"
    displayHint = "255a"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(1, 255),
    )

    if mibBuilder.loadTexts:
        description = "Represents a DNS domain name. The name SHOULD be fully qualified whenever possible. The corresponding InetAddressType is dns(16). The DESCRIPTION clause of InetAddress objects that may have InetAddressDNS values MUST fully describe how (and when) these names are to be resolved to IP addresses. The resolution of an InetAddressDNS value may require to query multiple DNS records (e.g., A for IPv4 and AAAA for IPv6). The order of the resolution process and which DNS record takes precedence depends on the configuration of the resolver. This textual convention SHOULD NOT be used directly in object definitions, as it restricts addresses to a specific format. However, if it is used, it MAY be used either on its own or in conjunction with InetAddressType, as a pair."


# IMPORTANT: customization
class InetAddress(TextualConvention, OctetString):
    status = "current"
    subtypeSpec = OctetString.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueSizeConstraint(0, 255),
    )

    if mibBuilder.loadTexts:
        description = "Denotes a generic Internet address. An InetAddress value is always interpreted within the context of an InetAddressType value. Every usage of the InetAddress textual convention is required to specify the InetAddressType object that provides the context. It is suggested that the InetAddressType object be logically registered before the object(s) that use the InetAddress textual convention, if they appear in the same logical row. The value of an InetAddress object must always be consistent with the value of the associated InetAddressType object. Attempts to set an InetAddress object to a value inconsistent with the associated InetAddressType must fail with an inconsistentValue error. When this textual convention is used as the syntax of an index object, there may be issues with the limit of 128 sub-identifiers specified in SMIv2, STD 58. In this case, the object definition MUST include a 'SIZE' clause to limit the number of potential instance sub-identifiers; otherwise the applicable constraints MUST be stated in the appropriate conceptual row DESCRIPTION clauses, or in the surrounding documentation if there is no single DESCRIPTION clause that is appropriate."

    # IMPORTANT: customization
    typeMap = {
        InetAddressType.namedValues.getValue("ipv4"): InetAddressIPv4(),
        InetAddressType.namedValues.getValue("ipv6"): InetAddressIPv6(),
        InetAddressType.namedValues.getValue("ipv4z"): InetAddressIPv4z(),
        InetAddressType.namedValues.getValue("ipv6z"): InetAddressIPv6z(),
        InetAddressType.namedValues.getValue("dns"): InetAddressDNS(),
    }

    @classmethod
    def cloneFromName(cls, value, impliedFlag, parentRow, parentIndices):
        for parentIndex in reversed(parentIndices):
            if isinstance(parentIndex, InetAddressType):
                try:
                    return parentRow.setFromName(
                        cls.typeMap[int(parentIndex)], value, impliedFlag, parentIndices
                    )
                except KeyError:
                    pass

        raise error.SmiError(
            f"{cls.__name__} object encountered without preceding InetAddressType-like index: {value!r}"
        )

    def cloneAsName(self, impliedFlag, parentRow, parentIndices):
        for parentIndex in reversed(parentIndices):
            if isinstance(parentIndex, InetAddressType):
                try:
                    return parentRow.getAsName(
                        self.typeMap[int(parentIndex)].clone(
                            self.asOctets().decode("ascii")
                        ),
                        impliedFlag,
                        parentIndices,
                    )
                except KeyError:
                    pass

        raise error.SmiError(
            f"{self.__class__.__name__} object encountered without preceding InetAddressType-like index: {self!r}"
        )


class InetAddressPrefixLength(TextualConvention, Unsigned32):
    status = "current"
    displayHint = "d"
    subtypeSpec = Unsigned32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueRangeConstraint(0, 2040),
    )

    if mibBuilder.loadTexts:
        description = "Denotes the length of a generic Internet network address prefix. A value of n corresponds to an IP address mask that has n contiguous 1-bits from the most significant bit (MSB), with all other bits set to 0. An InetAddressPrefixLength value is always interpreted within the context of an InetAddressType value. Every usage of the InetAddressPrefixLength textual convention is required to specify the InetAddressType object that provides the context. It is suggested that the InetAddressType object be logically registered before the object(s) that use the InetAddressPrefixLength textual convention, if they appear in the same logical row. InetAddressPrefixLength values larger than the maximum length of an IP address for a specific InetAddressType are treated as the maximum significant value applicable for the InetAddressType. The maximum significant value is 32 for the InetAddressType 'ipv4(1)' and 'ipv4z(3)' and 128 for the InetAddressType 'ipv6(2)' and 'ipv6z(4)'. The maximum significant value for the InetAddressType 'dns(16)' is 0. The value zero is object-specific and must be defined as part of the description of any object that uses this syntax. Examples of the usage of zero might include situations where the Internet network address prefix is unknown or does not apply. The upper bound of the prefix length has been chosen to be consistent with the maximum size of an InetAddress."


class InetPortNumber(TextualConvention, Unsigned32):
    status = "current"
    displayHint = "d"
    subtypeSpec = Unsigned32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        ValueRangeConstraint(0, 65535),
    )

    if mibBuilder.loadTexts:
        description = "Represents a 16 bit port number of an Internet transport layer protocol. Port numbers are assigned by IANA. A current list of all assignments is available from <http://www.iana.org/>. The value zero is object-specific and must be defined as part of the description of any object that uses this syntax. Examples of the usage of zero might include situations where a port number is unknown, or when the value zero is used as a wildcard in a filter."
    if mibBuilder.loadTexts:
        reference = "STD 6 (RFC 768), STD 7 (RFC 793) and RFC 2960"


class InetAutonomousSystemNumber(TextualConvention, Unsigned32):
    status = "current"
    displayHint = "d"
    if mibBuilder.loadTexts:
        description = "Represents an autonomous system number that identifies an Autonomous System (AS). An AS is a set of routers under a single technical administration, using an interior gateway protocol and common metrics to route packets within the AS, and using an exterior gateway protocol to route packets to other ASes'. IANA maintains the AS number space and has delegated large parts to the regional registries. Autonomous system numbers are currently limited to 16 bits (0..65535). There is, however, work in progress to enlarge the autonomous system number space to 32 bits. Therefore, this textual convention uses an Unsigned32 value without a range restriction in order to support a larger autonomous system number space."
    if mibBuilder.loadTexts:
        reference = "RFC 1771, RFC 1930"


class InetScopeType(TextualConvention, Integer32):
    status = "current"
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(1,
              2,
              3,
              4,
              5,
              8,
              14)
        )
    )
    namedValues = NamedValues(
        *(("interfaceLocal", 1),
          ("linkLocal", 2),
          ("subnetLocal", 3),
          ("adminLocal", 4),
          ("siteLocal", 5),
          ("organizationLocal", 8),
          ("global", 14))
    )

    if mibBuilder.loadTexts:
        description = "Represents a scope type. This textual convention can be used in cases where a MIB has to represent different scope types and there is no context information, such as an InetAddress object, that implicitly defines the scope type. Note that not all possible values have been assigned yet, but they may be assigned in future revisions of this specification. Applications should therefore be able to deal with values not yet assigned."
    if mibBuilder.loadTexts:
        reference = "RFC 3513"


class InetZoneIndex(TextualConvention, Unsigned32):
    status = "current"
    displayHint = "d"
    if mibBuilder.loadTexts:
        description = "A zone index identifies an instance of a zone of a specific scope. The zone index MUST disambiguate identical address values. For link-local addresses, the zone index will typically be the interface index (ifIndex as defined in the IF-MIB) of the interface on which the address is configured. The zone index may contain the special value 0, which refers to the default zone. The default zone may be used in cases where the valid zone index is not known (e.g., when a management application has to write a link-local IPv6 address without knowing the interface index value). The default zone SHOULD NOT be used as an easy way out in cases where the zone index for a non-global IPv6 address is known."
    if mibBuilder.loadTexts:
        reference = "RFC4007"


class InetVersion(TextualConvention, Integer32):
    status = "current"
    subtypeSpec = Integer32.subtypeSpec
    subtypeSpec += ConstraintsUnion(
        SingleValueConstraint(
            *(0,
              1,
              2)
        )
    )
    namedValues = NamedValues(
        *(("unknown", 0),
          ("ipv4", 1),
          ("ipv6", 2))
    )

    if mibBuilder.loadTexts:
        description = "A value representing a version of the IP protocol. unknown(0) An unknown or unspecified version of the IP protocol. ipv4(1) The IPv4 protocol as defined in RFC 791 (STD 5). ipv6(2) The IPv6 protocol as defined in RFC 2460. Note that this textual convention SHOULD NOT be used to distinguish different address types associated with IP protocols. The InetAddressType has been designed for this purpose."
    if mibBuilder.loadTexts:
        reference = "RFC 791, RFC 2460"


# MIB Managed Objects in the order of their OIDs


# Managed Objects groups


# Notification objects


# Notifications groups


# Agent capabilities


# Module compliance


# Export all MIB objects to the MIB builder

mibBuilder.export_symbols(
    "INET-ADDRESS-MIB",
    **{"InetAddressType": InetAddressType,
       "InetAddress": InetAddress,
       "InetAddressIPv4": InetAddressIPv4,
       "InetAddressIPv6": InetAddressIPv6,
       "InetAddressIPv4z": InetAddressIPv4z,
       "InetAddressIPv6z": InetAddressIPv6z,
       "InetAddressDNS": InetAddressDNS,
       "InetAddressPrefixLength": InetAddressPrefixLength,
       "InetPortNumber": InetPortNumber,
       "InetAutonomousSystemNumber": InetAutonomousSystemNumber,
       "InetScopeType": InetScopeType,
       "InetZoneIndex": InetZoneIndex,
       "InetVersion": InetVersion,
       "inetAddressMIB": inetAddressMIB,
       "PYSNMP_MODULE_ID": inetAddressMIB}
)



import asyncio
from datetime import datetime
import pytest

from pysnmp.hlapi.v3arch.asyncio import *
from pysnmp.proto.errind import RequestTimedOut
from pysnmp.proto.rfc1905 import errorStatus as pysnmp_errorStatus

from tests.agent_context import AGENT_PORT, AgentContextManager


@pytest.mark.asyncio
async def test_v1_get():
    async with AgentContextManager():
        with SnmpEngine() as snmpEngine:
            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpEngine,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(("localhost", AGENT_PORT)),
                ContextData(),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysDescr.0"
            assert varBinds[0][1].prettyPrint().startswith("PySNMP engine version")
            assert isinstance(varBinds[0][1], OctetString)


@pytest.mark.asyncio
async def test_v1_get_ipv6():
    async with AgentContextManager(enable_ipv6=True):
        with SnmpEngine() as snmpEngine:
            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpEngine,
                CommunityData("public", mpModel=0),
                await Udp6TransportTarget.create(("localhost", AGENT_PORT)),
                ContextData(),
                ObjectType(ObjectIdentity("SNMPv2-MIB", "sysDescr", 0)),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert len(varBinds) == 1
            assert varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysDescr.0"
            assert varBinds[0][1].prettyPrint().startswith("PySNMP engine version")
            assert isinstance(varBinds[0][1], OctetString)


@pytest.mark.asyncio
async def test_v1_get_timeout_invalid_target():
    with SnmpEngine() as snmpEngine:

        async def run_get():
            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpEngine,
                CommunityData("community_string"),
                await UdpTransportTarget.create(("1.2.3.4", 161), timeout=1, retries=0),
                ContextData(),
                ObjectType(ObjectIdentity("1.3.6.1.4.1.60069.9.1.0")),
            )
            assert isinstance(errorIndication, RequestTimedOut)
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 0

        start = datetime.now()
        try:
            await asyncio.wait_for(run_get(), timeout=3)
            end = datetime.now()
            elapsed_time = (end - start).total_seconds()
            assert (
                elapsed_time >= 1 and elapsed_time <= 3
            )  # transport timeout is 1 second
        except asyncio.TimeoutError:
            assert False, "Test case timed out"


@pytest.mark.asyncio
async def test_v1_get_slow_object():
    async with AgentContextManager(enable_custom_objects=True):
        with SnmpEngine() as snmpEngine:

            async def run_get():
                errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                    snmpEngine,
                    CommunityData("public", mpModel=0),
                    await UdpTransportTarget.create(
                        ("localhost", AGENT_PORT),
                        timeout=1,
                        retries=0,  # TODO: why this timeout did not work?
                    ),
                    ContextData(),
                    ObjectType(ObjectIdentity("1.3.6.1.4.1.60069.9.1.0")),
                )
                assert errorIndication is None
                assert errorStatus == 0
                assert len(varBinds) == 1

            start = datetime.now()
            try:
                await asyncio.wait_for(run_get(), timeout=3)
                end = datetime.now()
                elapsed_time = (end - start).total_seconds()
                assert (
                    elapsed_time >= 2 and elapsed_time <= 3
                )  # 2 seconds is the delay for the object
            except asyncio.TimeoutError:
                assert False, "Test case timed out"


@pytest.mark.asyncio
async def test_v1_get_no_access_object():
    async with AgentContextManager(enable_custom_objects=True):
        with SnmpEngine() as snmpEngine:
            errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
                snmpEngine,
                CommunityData("public", mpModel=0),
                await UdpTransportTarget.create(
                    ("localhost", AGENT_PORT), timeout=1, retries=0
                ),
                ContextData(),
                ObjectType(ObjectIdentity("1.3.6.1.4.1.60069.9.3")),
            )
            assert errorIndication is None
            assert (
                errorStatus.prettyPrint() == "noSuchName"
            )  # v1 does not have noAccess



import pytest
from pysnmp.hlapi.v3arch.asyncio import *
from pysnmp.smi import builder, compiler, view
from tests.manager_context import MANAGER_PORT, ManagerContextManager


@pytest.mark.asyncio
async def test_send_v3_inform():
    async with ManagerContextManager():
        with SnmpEngine() as snmpEngine:
            (
                errorIndication,
                errorStatus,
                errorIndex,
                varBinds,
            ) = await send_notification(
                snmpEngine,
                UsmUserData("usr-md5-des", "authkey1", "privkey1"),
                await UdpTransportTarget.create(("localhost", MANAGER_PORT)),
                ContextData(),
                "inform",
                NotificationType(ObjectIdentity("1.3.6.1.6.3.1.1.5.2"))
                .load_mibs("SNMPv2-MIB")
                .add_varbinds(("1.3.6.1.2.1.1.1.0", OctetString("my system"))),
            )

            assert errorIndication is None
            assert errorStatus == 0
            assert errorIndex == 0
            assert len(varBinds) == 3
            assert (
                varBinds[0][0].prettyPrint() == "SNMPv2-MIB::sysUpTime.0"
            )  # IMPORTANT: MIB is needed to resolve this name
            assert varBinds[1][0].prettyPrint() == "SNMPv2-MIB::snmpTrapOID.0"
            assert varBinds[2][0].prettyPrint() == "SNMPv2-MIB::sysDescr.0"
            isinstance(varBinds[0][1], TimeTicks)
            isinstance(varBinds[1][1], ObjectIdentifier)
            isinstance(varBinds[2][1], OctetString)



"""
Custom ASN.1 MIB path
+++++++++++++++++++++

Send SNMP GET request using the following options:

* with SNMPv2c, community 'public'
* over IPv4/UDP
* to an Agent at demo.pysnmp.com:161
* for IF-MIB::ifInOctets.1 MIB object
* pass non-default ASN.1 MIB source to MIB compiler

Functionally similar to:

| $ snmpget -v2c -c public -M /usr/share/snmp demo.pysnmp.com IF-MIB::ifInOctets.1

"""  #
import asyncio
from pysnmp.hlapi.v3arch.asyncio import *


async def run():
    snmpEngine = SnmpEngine()
    errorIndication, errorStatus, errorIndex, varBinds = await get_cmd(
        snmpEngine,
        CommunityData("public"),
        await UdpTransportTarget.create(("demo.pysnmp.com", 161)),
        ContextData(),
        ObjectType(
            ObjectIdentity("IF-MIB", "ifInOctets", 1).add_asn1_mib_source(
                "file:///usr/share/snmp", "https://mibs.pysnmp.com/asn1/@mib@"
            )
        ),
    )

    if errorIndication:
        print(errorIndication)
    elif errorStatus:
        print(
            "{} at {}".format(
                errorStatus.prettyPrint(),
                errorIndex and varBinds[int(errorIndex) - 1][0] or "?",
            )
        )
    else:
        for varBind in varBinds:
            print(" = ".join([x.prettyPrint() for x in varBind]))


asyncio.run(run())



#
# This file is part of pysnmp software.
#
# Copyright (c) 2005-2020, Ilya Etingof <etingof@gmail.com>
# License: https://www.pysnmp.com/pysnmp/license.html
#
# PySNMP MIB module PYSNMP-SOURCE-MIB (https://www.pysnmp.com/pysnmp)
# ASN.1 source http://mibs.pysnmp.com:80/asn1/PYSNMP-SOURCE-MIB
# Produced by pysmi-0.1.3 at Mon Apr 17 11:46:02 2017
# On host grommit.local platform Darwin version 16.4.0 by user ilya
# Using Python version 3.4.2 (v3.4.2:ab2c023a9432, Oct  5 2014, 20:42:22)
#
Integer, OctetString, ObjectIdentifier = mibBuilder.import_symbols(
    "ASN1", "Integer", "OctetString", "ObjectIdentifier"
)
(NamedValues,) = mibBuilder.import_symbols("ASN1-ENUMERATION", "NamedValues")
(
    SingleValueConstraint,
    ValueRangeConstraint,
    ConstraintsIntersection,
    ValueSizeConstraint,
    ConstraintsUnion,
) = mibBuilder.import_symbols(
    "ASN1-REFINEMENT",
    "SingleValueConstraint",
    "ValueRangeConstraint",
    "ConstraintsIntersection",
    "ValueSizeConstraint",
    "ConstraintsUnion",
)
(pysnmpModuleIDs,) = mibBuilder.import_symbols("PYSNMP-MIB", "pysnmpModuleIDs")
(snmpTargetAddrEntry,) = mibBuilder.import_symbols(
    "SNMP-TARGET-MIB", "snmpTargetAddrEntry"
)
NotificationGroup, ModuleCompliance = mibBuilder.import_symbols(
    "SNMPv2-CONF", "NotificationGroup", "ModuleCompliance"
)
(
    ModuleIdentity,
    iso,
    MibScalar,
    MibTable,
    MibTableRow,
    MibTableColumn,
    Gauge32,
    NotificationType,
    IpAddress,
    MibIdentifier,
    Unsigned32,
    Counter32,
    ObjectIdentity,
    Counter64,
    Bits,
    Integer32,
    TimeTicks,
) = mibBuilder.import_symbols(
    "SNMPv2-SMI",
    "ModuleIdentity",
    "iso",
    "MibScalar",
    "MibTable",
    "MibTableRow",
    "MibTableColumn",
    "Gauge32",
    "NotificationType",
    "IpAddress",
    "MibIdentifier",
    "Unsigned32",
    "Counter32",
    "ObjectIdentity",
    "Counter64",
    "Bits",
    "Integer32",
    "TimeTicks",
)
TextualConvention, DisplayString, TAddress = mibBuilder.import_symbols(
    "SNMPv2-TC", "TextualConvention", "DisplayString", "TAddress"
)
pysnmpSourceMIB = ModuleIdentity((1, 3, 6, 1, 4, 1, 20408, 3, 1, 8))
if mibBuilder.loadTexts:
    pysnmpSourceMIB.setRevisions(
        (
            "2017-04-14 00:00",
            "2015-01-16 00:00",
        )
    )
if mibBuilder.loadTexts:
    pysnmpSourceMIB.setLastUpdated("201704140000Z")
if mibBuilder.loadTexts:
    pysnmpSourceMIB.setOrganization("The PySNMP Project")
if mibBuilder.loadTexts:
    pysnmpSourceMIB.setContactInfo(
        "E-mail: LeXtudio Inc. <support@lextudio.com> GitHub: https://github.com/lextudio/pysnmp"
    )
if mibBuilder.loadTexts:
    pysnmpSourceMIB.setDescription(
        "This MIB module defines implementation specific objects that provide variable source transport endpoints feature to SNMP Engine and Standard SNMP Applications."
    )
pysnmpSourceMIBObjects = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3, 1, 8, 1))
pysnmpSourceMIBConformance = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3, 1, 8, 2))
snmpSourceAddrTable = MibTable(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 8, 1, 1),
)
if mibBuilder.loadTexts:
    snmpSourceAddrTable.setStatus("current")
if mibBuilder.loadTexts:
    snmpSourceAddrTable.setDescription(
        "A table of transport addresses to be used as a source in the generation of SNMP messages. This table contains additional objects for the SNMP-TRANSPORT-ADDRESS::snmpSourceAddressTable."
    )
snmpSourceAddrEntry = MibTableRow(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 8, 1, 1, 1),
)
snmpTargetAddrEntry.registerAugmentions(("PYSNMP-SOURCE-MIB", "snmpSourceAddrEntry"))
snmpSourceAddrEntry.setIndexNames(*snmpTargetAddrEntry.getIndexNames())
if mibBuilder.loadTexts:
    snmpSourceAddrEntry.setStatus("current")
if mibBuilder.loadTexts:
    snmpSourceAddrEntry.setDescription(
        "A transport address to be used as a source in the generation of SNMP operations. An entry containing additional management information applicable to a particular target."
    )
snmpSourceAddrTAddress = MibTableColumn(
    (1, 3, 6, 1, 4, 1, 20408, 3, 1, 8, 1, 1, 1, 1), TAddress()
).setMaxAccess("read-create")
if mibBuilder.loadTexts:
    snmpSourceAddrTAddress.setStatus("current")
if mibBuilder.loadTexts:
    snmpSourceAddrTAddress.setDescription(
        "This object contains a transport address. The format of this address depends on the value of the snmpSourceAddrTDomain object."
    )
pysnmpSourceMIBCompliances = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3, 1, 8, 2, 1))
pysnmpSourceMIBGroups = MibIdentifier((1, 3, 6, 1, 4, 1, 20408, 3, 1, 8, 2, 2))
mibBuilder.export_symbols(
    "PYSNMP-SOURCE-MIB",
    pysnmpSourceMIBConformance=pysnmpSourceMIBConformance,
    pysnmpSourceMIB=pysnmpSourceMIB,
    snmpSourceAddrTable=snmpSourceAddrTable,
    snmpSourceAddrEntry=snmpSourceAddrEntry,
    pysnmpSourceMIBGroups=pysnmpSourceMIBGroups,
    PYSNMP_MODULE_ID=pysnmpSourceMIB,
    snmpSourceAddrTAddress=snmpSourceAddrTAddress,
    pysnmpSourceMIBObjects=pysnmpSourceMIBObjects,
    pysnmpSourceMIBCompliances=pysnmpSourceMIBCompliances,
)
